﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;


namespace LockDownSvc
{
    /// <summary>
    /// Store user time configuration information such as time Span and time Left.
    /// 
    /// 
    /// TODO: Add Idle span,left,logout
    ///       network span,left
    ///       expand span,left to more complex day,week,calendar
    /// 
    /// Author: Dennis Lang
    /// LockDown service - Dec 2010
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public class Config
    {
        public const string AdminUserKeyStr = "admin-users";
        public const string InternalKeyStr = "internal";

        #region ==== Construction
        public Config() 
        {
            Date = DateTime.Now;
            Left = Span = new TimeSpan(2, 0, 0);   // Default 2 hours per day.
            Elapsed = TimeSpan.Zero;
            User = WinFunc.UnknownUser;
        }

        /// <summary>
        /// Populate some statics used by all instances of Config.
        /// </summary>
        /// <param name="dataMgr"></param>
        /// <param name="baseDir"></param>
        public static void Initialize(DataMgr dataMgr, string baseDir)
        {
            DataManager = dataMgr;
            BaseDir = baseDir;
            LoadAdminUsers();
        }

        private static DataMgr DataManager { get; set; }
        private static string BaseDir { get; set; }

        public static  bool IsUnlimited(TimeSpan ts)
        {
            return (ts == TimeSpan.MaxValue);
        }
       
        #endregion

        #region ==== Main interface

        // ToDo - Expand to weekly schedule.

        /// <summary>
        /// Current user associated with this configuration.
        /// </summary>
        public string User { set; get; }

        /// <summary>
        /// Persistent date use to determine when to reset daily time left.
        /// </summary>
        public DateTime Date 
        {
            set { m_dateTime = value; }
            get { return m_dateTime; }
        }
        DateTime m_dateTime;

        /// <summary>
        /// Persistent time span used to reset time left when day changes.
        /// </summary>
        public TimeSpan Span { set; get; }

        /// <summary>
        /// Persistent time left, reset to 'Span' when day changes.
        /// </summary>
        public TimeSpan Left
        {
            get 
            {
                lock (this)     // thread safe.
                {
                    DateTime now = DateTime.Now;
                    if (now < Date)
                    {
                        // Odd - time has gone backwards, user must be messing with the clock.
                        Date = now;
                        return m_left;
                    }

                    DailyReset();
                    TimeSpan elapsed = (now - Date);

                    if (elapsed.TotalSeconds > 0)
                    {
                        // Ignore large time gaps, something is wrong like user is messing with clock.
                        if (elapsed.TotalMinutes < 5)
                        {
                            if (!IsUnlimited(m_left))
                                m_left -= elapsed;
                            Elapsed += elapsed;
                        }

                        Date = now;
                    }

                    return m_left; 
                }
            }

            set { m_left = value; Date = DateTime.Now; }
        }
        private TimeSpan m_left = new TimeSpan(0);

        /// <summary>
        /// Time elapsed while logged in.
        /// </summary>
        public TimeSpan Elapsed { set; get; }

        #endregion

        #region ==== String manipulation

        /// <summary>
        /// Return string version of time left (handle special cases like unlimited).
        /// </summary>
        public string LeftString()
        {
            return IsUnlimited(Left) ? "Unlimited" : ToHHMMSS(Left);
        }

        /// <summary>
        /// Return string version of time span (handle special cases like unlimited).
        /// </summary>
        /// <returns></returns>
        public string SpanString()
        {
            return IsUnlimited(Span) ? "Unlimited" : ToHHMMSS(Span);
        }

        public string ToHHMMSS(TimeSpan ts)
        {
            if (ts.TotalMinutes < 0)
                return "0:0:0";
            return string.Format("{0}:{1:00}:{2:00}", ts.Hours, ts.Minutes, ts.Seconds);
        }

        /// <summary>
        /// Parse string to 'span' value.
        /// </summary>
        /// <param name="value"></param>
        /// <returns>true if valid string</returns>
        public bool ParseSpan(string value)
        {
            try
            {
                this.Span = ((value.ToLower() == "unlimited") ?
                    TimeSpan.MaxValue : TimeSpan.Parse(value));
                return true;
            }
            catch { }
            return false;
        }

        /// <summary>
        /// Parse string to 'left' value.
        /// </summary>
        /// <param name="value"></param>
        /// <returns>true if valid string</returns>
        public bool ParseLeft(string value)
        {
            try
            {
                this.Left = ((value.ToLower() == "unlimited") ?
                    TimeSpan.MaxValue : TimeSpan.Parse(value));
                return true;
            }
            catch { }
            return false;
        }

        /// <summary>
        /// Return string representation of configuration.
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return
                "User=" + User
                + ", Date=" + Date.ToString("G")
                + ", Span=" + SpanString()
                + ", Left=" + LeftString()
                + ", Elapsed=" + ToHHMMSS(Elapsed);
        }

        #endregion

        #region ==== Misc methods
        /// <summary>
        /// If day has changed, reset time 'Left' time to time 'Span'.
        /// </summary>
        /// <returns></returns>
        internal bool DailyReset()
        {
            DateTime now = DateTime.Now;
            if (now > Date && now.DayOfYear != Date.DayOfYear)
            {
                Date = now;
                Left = Span;
                return true;
            }

            return false;
        }

        /// <summary>
        /// Set time 'Span' to "unlimited". 
        /// </summary>
        public void SetToUnlimited()
        {
            Left = Span = TimeSpan.MaxValue;
        }

        /// <summary>
        /// Return true if authorized LockDown administrator account.
        /// If no admin user set, then everyone is admin (handy during startup condition).
        /// </summary>
        /// <returns></returns>
        public static bool IsAdmin(string user)
        {
            return (AdminUsers.Length == 0) || AdminUsers.Contains(user + ';');
        }

        /// <summary>
        /// Return list of admin users as user1;user2;
        /// </summary>
        public static string AdminUsers 
        {
            get { return s_adminUsers; }
            set 
            {
                if (s_adminUsers != value)
                {
                    s_adminUsers = value;
                    SaveAdminUsers();
                }
            }
        }

        public static DataMgr InternalDataMgr()
        {
            return DataManager.SetChild(InternalKeyStr);
        }

        /// <summary>
        /// Save persistent Admin user list.
        /// </summary>
        public static bool SaveAdminUsers()
        {
            try
	        {
                DataMgr dataMgr = InternalDataMgr();
	            dataMgr.SetString(AdminUserKeyStr, AdminUsers);
	            dataMgr.Apply();
	                return true;
            }
            catch (Exception ex)
            {
                // NonAdmin cannot save Admin users.
            }

            return false;
        }

        private static void LoadAdminUsers()
        {
            string adminUsers = InternalDataMgr().GetString(AdminUserKeyStr);
            if (adminUsers == null)
                adminUsers = string.Empty;
            Config.AdminUsers = adminUsers;
        }

        private static string s_adminUsers;
        #endregion

        #region ==== Persistent interfaces File, Load, Save

        /// <summary>
        /// Load configuration from disk for user.
        /// </summary>
        /// <returns>true if successful</returns>
        public bool LoadFor(string user, DataMgr dataMgr)
        {
            try
            {
                if (dataMgr.HasChild(user))
                {
                    dataMgr.SetChild(user);
                    ParseSpan(dataMgr.GetString("span"));
                    this.Date = DateTime.Parse(dataMgr.GetString("date"));
                    if (!this.DailyReset())
                        ParseLeft(dataMgr.GetString("left"));   // Warning - sets date to 'now'.
                    this.User = user;
                    this.Elapsed = TimeSpan.Zero;

                    return true;
                }
                else
                {
                    Log.LogEvent("No Cfg For:" + user);
                }
            }
            catch (Exception ex)
            {
                Log.LogEvent("LoadCfg failed " + ex.Message);
            }

            this.Span = this.Left = new TimeSpan(2, 0, 0);
            this.User = user;
            this.Elapsed = TimeSpan.Zero;
            return false;
        }

        /// <summary>
        /// Save Cfg to file.
        /// </summary>
        public void SaveFor(string user, DataMgr dataMgr)
        {
            dataMgr.SetChild(user);
            dataMgr.SetString("date", this.Date.ToString());
            dataMgr.SetString("span", this.SpanString());
            dataMgr.SetString("left", this.LeftString());
            dataMgr.SetString("elapsed", ToHHMMSS(this.Elapsed));       // saved but not loaded.
            dataMgr.Apply();
        }

        #endregion
    }
}
