﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.ServiceProcess;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Media;

namespace LockDownSvc
{
    /// <summary>
    /// Parental Control running as a window service to limit daily user time.
    /// 
    /// Author: Dennis Lang with some support routines by Pascal Ganaye
    /// LockDown service - Dec 2010
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public class LockDownBg : LockDown
    {
        private System.Timers.Timer m_sysTimer = new System.Timers.Timer(1000);
        private int m_Seq = -1;
        private int m_guiPid = -1;
        private DateTime m_logOnAt = DateTime.MaxValue;
        private int m_warningCount = 0;

        #region ==== Construction
        public LockDownBg()
        {
            Action = Actions.eInit;      

            // Start timer.
            m_sysTimer.Elapsed += new System.Timers.ElapsedEventHandler(Timer_tick);
            m_sysTimer.Enabled = true;
        }
        #endregion

        #region ==== Manage user account state changes (Start, LogOn, LogOff, Stop)
        public void Start()
        {
            Log.LogEvent("Start");
        }

        #region ==== Account state changes
        private SensAdvisor m_Advisor;
        private void InitializeAccountMonitor()
        {
            if (m_Advisor == null)
            {
                // Monitor account log on, off, reconnect, disconnect actions.
                try
                {
                    m_Advisor = new SensAdvisor();
                    m_Advisor.OnShellStarted += this.PostShell;
                    m_Advisor.OnLogoff += this.LogOff;
                    m_Advisor.OnSessionDisconnected += this.LogOff;
                    m_Advisor.OnLogon += this.LogOn;
                    m_Advisor.OnSessionReconnected += this.LogOn;
                }
                catch (Exception ex)
                {
                    Log.LogEvent("LockDown - advisor failed");
                    m_Advisor = null;
                }
            }
        }

        public void PostShell(object sender, SensLogon2EventArgs e)
        {
        }

        public void LogOff(object sender, SensLogon2EventArgs e)
        {
            LogOff(e.Username);
        }

        public void LogOn(object sender, SensLogon2EventArgs e)
        {
            LogOn(e.Username);
        }
        #endregion


        public void LogOn(string user)
        {
            m_logOnAt = DateTime.Now;
            m_warningCount = 0;

            WinFunc.ForegroundUser = user;
            if (Action != Actions.eIdle)
                Log.LogEvent("LogOn, lost action=" + Action.ToString());
            Action = Actions.eMonitor;
            LoadConfig();
            Log.LogMessage("LogOn, " + Configuration.ToString());
        }

        public void LogOff(string user)
        {
            WinFunc.ForegroundUser = user;
            Action = Actions.eIdle;
            Log.LogMessage("LogOff, " + Configuration.ToString());
            SaveConfig();

            try
            {
                if (m_guiPid > 1)
                    Process.GetProcessById(m_guiPid).Kill();
            }
            catch { }
            m_guiPid = -1;
        }

        public void Stop()
        {
            Action = Actions.eIdle;
            Log.LogEvent("Stop");

            try
            {
                if (m_guiPid > 1)
                    Process.GetProcessById(m_guiPid).Kill();
            }
            catch { }
            m_guiPid = -1;
        }
        #endregion

        #region ===== Timer and Action commands
        public static Fg2BgCommands ReadBgCommand(ref int lastSeq)
        {
            return (Fg2BgCommands)ReadCommand(s_cmdBgOff, ref lastSeq);
        }
        public static void WriteFgCommand(Bg2FgCommands cmd)
        {
            WriteCommand((byte)cmd, s_cmdFgOff);
        }

        /// <summary>
        /// Asyncrhonouse actions - allows threads to communication to request an action.
        /// </summary>
        public enum Actions { eInit, eIdle, eMonitor, ePause, eContinue };
        public Actions Action { get; set; }

        /// <summary>
        /// Timer to execute asynchronous actions and check if time to log user off.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        internal void Timer_tick(object sender, System.Timers.ElapsedEventArgs e)
        {
            WriteFgCommand(Bg2FgCommands.eHello);

            switch (ReadBgCommand(ref m_Seq))
            {
                case Fg2BgCommands.eLoadCfg:        // when Config editor applies changes.
                    LoadConfig();
                    Log.LogMessage("LoadCfg, " + Configuration.ToString());
                    // LoadConfigFromShdMem();
                    break;
                case Fg2BgCommands.eLogOff:
                    Log.LogEvent("bg got logoff");
                    ForceLogOff();
                    break;
                case Fg2BgCommands.eEndSession:
                    Log.LogMessage("EndSession, " + Configuration.ToString());
                    Log.LogEvent("EndSession, " + Configuration.ToString());

                    Action = Actions.eIdle;
                    Log.LogMessage("LogOff, " + Configuration.ToString());
                    SaveConfig();
                    break;
                case Fg2BgCommands.eHello:
                    m_lastHello = DateTime.Now;
                    break;
            }

            switch (Action)
            {
                case Actions.eInit:
                    m_sysTimer.Enabled = false;
                    Action = Actions.eMonitor;
                    // Action = Actions.eIdle;
                    WinFunc.ForegroundUser = WinFunc.GetUser();
                    InitializeLockDown();
                    InitializeAccountMonitor();
                    // LoadAllCfgsToShdMem();
                    Configuration = new Config();
                    LoadConfig();
                    m_sysTimer.Enabled = true;
                    return;                  

                case Actions.eIdle:
                    return;

                case Actions.eMonitor:
                    LaunchGui();    // Launch GUI if not already running or if died.
                    SaveConfig();   // Periodically save Configuration.

                    // Check if time to force user off, give a few minutes of curtise if just logging in.
                    if (Configuration.Left.TotalSeconds <= 0 && (DateTime.Now - m_logOnAt).TotalMinutes > 2)
                    {
                        PlaySound();
                        ForceLogOff();      
                    }
                    else if (Configuration.Left.TotalMinutes < 5)
                    {
                        if (m_warningCount++ == 0)
                        {
                            // Tell foreground almost time to logOff
                            PlaySound();
                            WriteFgCommand(Bg2FgCommands.eAlmostToast);
                        }
                    }
                    else
                    {
                        m_warningCount = 0;
                    }
                    break;

                case Actions.ePause:
                    return;

                case Actions.eContinue:
                    Configuration.Date = DateTime.Now;
                    Action = Actions.eMonitor;
                    break;
            }
        }

        DateTime m_lastGuiCheck = DateTime.MinValue;
        DateTime m_lastHello = DateTime.Now;
        private void LaunchGui()
        {
            TimeSpan elapsed = DateTime.Now - m_lastGuiCheck;
            if (elapsed.TotalSeconds > 5)
            {
                m_lastGuiCheck = DateTime.Now;

                if (m_guiPid == -1)
                {
                    // Will fail to launch if 'explorer' is not found for 'user'.
                    m_guiPid = RunAs.Launch(Application.ExecutablePath + " -gui", WinFunc.ForegroundUser);
                    Log.LogEvent("Launch, guiPid=" + m_guiPid.ToString());
                }
            }

            elapsed = DateTime.Now - m_lastHello;
            if (elapsed.TotalSeconds > 5)
            {
                try
                {
                    // Verify GUI process is still running,
                    m_guiPid = Process.GetProcessById(m_guiPid).Id;
                }
                catch
                {
                    m_guiPid = -1;
                }
            }
        }

        /// <summary>
        /// Force user to log off.
        /// </summary>
        private void ForceLogOff()
        {
            Log.LogEvent("ForceLogOff, " + Configuration.ToString());
            SaveConfig();
            WriteFgCommand(Bg2FgCommands.eLogOff);

#if true
            // Background
            WinExitWindows.LogOffInteractiveUser();
#else
            // Foreground 
            WinExitWindows.ShutDown(
                  WinExitWindows.ShutDownMethod.LogOff | WinExitWindows.ShutDownMethod.ForceIfHung,
                  WinExitWindows.ShutDownReason.MajorOperatingSystem | WinExitWindows.ShutDownReason.FlagPlanned);
#endif
        }

        SoundPlayer m_player;
        DateTime m_lastSound = DateTime.MinValue;
        private void PlaySound()
        {
#if false
            TimeSpan elapsed = DateTime.Now - m_lastSound;

            if (elapsed.TotalSeconds > 30)
            {
                m_lastSound = DateTime.Now;

                try
                {
                    if (m_player == null)
                        m_player = new SoundPlayer();
                    m_player.Stream = Properties.Resources.sonar;
                    m_player.Play();
                }
                catch (Exception ex)
                {
                    System.Diagnostics.Debug.WriteLine(ex.Message);
                }
            }
#endif
        }

        #endregion

        #region ==== User Configuration
        public Config Configuration  { get; set; }
        private DateTime m_lastCfgSave = DateTime.MinValue;

        /// <summary>
        /// Load configuration.
        /// </summary>
        private void LoadConfig()
        {
            if (!Configuration.LoadFor(WinFunc.ForegroundUser, DataManager))
            {
                Configuration = new Config();
            }

            m_lastCfgSave = DateTime.Now;
        }

        /// <summary>
        /// Save configuration.
        /// </summary>
        private void SaveConfig()
        {
            TimeSpan elapsed = DateTime.Now - m_lastCfgSave;
            if (elapsed.TotalMinutes > 0)
            {
                m_lastCfgSave = DateTime.Now;

                try
                {
                    if (WinFunc.ForegroundUser == Configuration.User)
                    {
                        Configuration.SaveFor(WinFunc.ForegroundUser, DataManager);
                    }
                    else
                    {
                        Log.LogEvent("Skip save, wrong user");
                    }
                }
                catch (Exception ex)
                {
                    Log.LogEvent("Save Cfg error - " + ex.Message);
                }
            }
        }
        #endregion
    }
}
