﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Win32;
using System.IO;
using System.Security.AccessControl;
using System.Security.Principal;

namespace LockDownSvc
{
    /// <summary>
    /// Collection of classes to manage data using files or system registry.
    /// Data is stored in a hierarchy using (key, value) pairs.
    /// 
    /// Base class is DataMgr
    ///     RegMgr  - Store data in system registry.
    ///     FileMgr - Store data in files.
    /// 
    /// Author: Dennis Lang
    /// LockDown service - Dec 2010
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    /// 

    #region ==== Abstract base class
    public class DataMgr
    {
        public virtual DataMgr Clone()
        {
            throw new NotImplementedException();
        }

        public virtual string Root()
        {
            throw new NotImplementedException();
        }

        public virtual bool HasChild(string child)
        {
            throw new NotImplementedException();
        }

        public virtual string[] GetChildren()
        {
            throw new NotImplementedException();
        }

        public virtual DataMgr SetChild(string child)
        {
            throw new NotImplementedException();
        }



        public virtual bool HasKey(string key)
        {
            throw new NotImplementedException();
        }

        public virtual string[] GetKeys()
        {
            throw new NotImplementedException();
        }

        public virtual void DeleteKeys(string[] keys)
        {
            throw new NotImplementedException();
        }

        public virtual string GetString(string key)
        {
            throw new NotImplementedException();
        }
     
        public virtual void SetString(string key, string value)
        {
            throw new NotImplementedException();
        }

        public virtual void Apply()
        {
            throw new NotImplementedException();
        }

        public virtual void Cancel()
        {
            throw new NotImplementedException();
        }
    }
    #endregion

    #region ==== Registry data manager
    public class RegMgr : DataMgr
    {
        string m_rootKey;
        RegistryKey m_rootReg;
        RegistryKey m_childReg;

        public RegMgr(string rootKey)
        {
#if false
            RegistrySecurity rs = new RegistrySecurity();

            rs.AddAccessRule(new RegistryAccessRule("Everyone",
                RegistryRights.ReadKey,
                InheritanceFlags.ObjectInherit,
                PropagationFlags.InheritOnly,
                AccessControlType.Allow));

            rs.AddAccessRule(new RegistryAccessRule("Everyone",
                RegistryRights.WriteKey | RegistryRights.Delete | RegistryRights.ChangePermissions,
                InheritanceFlags.ObjectInherit,
                PropagationFlags.InheritOnly,
                AccessControlType.Deny));

            string currentUser = WindowsIdentity.GetCurrent().Name;
            rs.AddAccessRule(new RegistryAccessRule(currentUser,
               RegistryRights.WriteKey | RegistryRights.Delete | RegistryRights.ChangePermissions,
               InheritanceFlags.ObjectInherit,
               PropagationFlags.InheritOnly,
               AccessControlType.Allow));
#endif

            m_rootKey = rootKey;
            try
            {
                m_rootReg = Registry.LocalMachine.CreateSubKey(@"SOFTWARE\" + rootKey
                    /* , RegistryKeyPermissionCheck.ReadWriteSubTree */);
            }
            catch (Exception ex)
            {
                m_rootReg = Registry.LocalMachine.OpenSubKey(@"SOFTWARE\" + rootKey);
            }
            m_childReg = m_rootReg;
        }

        public RegMgr(string rootKey, RegistryKey baseReg)
        {
            m_rootKey = rootKey;
            m_rootReg = baseReg.CreateSubKey(rootKey);
            m_childReg = m_rootReg;
        }

        public RegMgr(string rootKey, RegistryKey rootReg, RegistryKey childReg)
        {
            m_rootKey = rootKey;
            m_rootReg = rootReg;
            m_childReg = childReg;
        }

        public override DataMgr Clone()
        {
            RegMgr regMgr = new RegMgr(m_rootKey, m_rootReg, m_childReg);
            return regMgr;
        }

        public override string Root()
        {
            return m_rootKey.ToString();
        }

        public override bool HasChild(string child)
        {
            return (-1 != Array.IndexOf(m_rootReg.GetSubKeyNames(), child));
        }

        public override string[] GetChildren()
        {
            return m_rootReg.GetSubKeyNames();
        }

        public override DataMgr SetChild(string child)
        {
            try
            {
            	m_childReg = m_rootReg.CreateSubKey(child);
            }
            catch (Exception ex)
            {
                m_childReg = m_rootReg.OpenSubKey(child);
            }
            return this;
        }



        public override bool HasKey(string key)
        {
            // return Array.Exists(m_userKey.GetValueNames(), element => element == key);
            return (-1 != Array.IndexOf(m_childReg.GetValueNames(), key));
        }

        public override string[] GetKeys()
        {
            return m_childReg.GetValueNames();
        }

        public override void DeleteKeys(string[] keys)
        {
            if (keys != null)
            {
                foreach (string key in keys)
                {
                    try
                    {
                        m_childReg.DeleteSubKey(key);
                    }
                    catch (Exception ex)
                    {
                        System.Diagnostics.Debug.WriteLine(ex.Message);
                    }
                }
            }
        }

        public override string GetString(string key)
        {
            return (string)m_childReg.GetValue(key);
        }

        public override void SetString(string key, string value)
        {
            m_childReg.SetValue(key, value);
        }

        public override void Apply()
        {
            // Data already applied
        }

        public override void Cancel()
        {
            throw new NotImplementedException();
        }
    }
    #endregion

    #region ==== File data manager
    /// <summary>
    /// File Manager - Store data as (key,value) pairs in a hierarchy of files.
    /// </summary>
    public class FileMgr : DataMgr
    {
        string m_rootDir;
        string m_childPath;
        Dictionary<string, string> m_values = null;

        /// <summary>
        /// Constructor - set root directory and file extension
        /// </summary>
        /// <param name="rootDir"></param>
        /// <param name="ext">ex: ".cfg" </param>
        public FileMgr(string rootDir, string ext)
        {
            m_rootDir = rootDir;
            Extension = ext;

            if (!Directory.Exists(m_rootDir))
            {
                Directory.CreateDirectory(m_rootDir);
            }
        }

        public string Extension { get; set; }

        public override string Root()
        {
            return m_rootDir;
        }

        public override DataMgr Clone()
        {
            FileMgr fileMgr = new FileMgr(m_rootDir, Extension);
            return fileMgr;
        }

        public override bool HasChild(string child)
        {
            return (-1 != Array.IndexOf(GetChildren(), child));
        }

        public override string[] GetChildren()
        {
            string[] files = Directory.GetFiles(m_rootDir, "*" + Extension);
            List<string> children = new List<string>();
            foreach (string file in files)
                children.Add(Path.GetFileNameWithoutExtension(file));
            return children.ToArray();
        }

        public override DataMgr SetChild(string child)
        {
            string childPath = System.IO.Path.Combine(m_rootDir, child + Extension);

            if (m_childPath != childPath)
            {
                m_childPath = childPath;
                m_values = null;
            }

            if (m_values == null)
                Load();
            return this;
        }

        public override bool HasKey(string key)
        {
            return m_values.ContainsKey(key);
        }

        public override string[] GetKeys()
        {
            List<string> keys = new List<string>();
            foreach (string s in m_values.Keys)
            {
                keys.Add(s);
            }

            return keys.ToArray();
        }

        public override void DeleteKeys(string[] keys)
        {
            if (keys != null)
            {
                foreach (string key in keys)
                {
                    m_values.Remove(key);
                }
            }
        }


        public override string GetString(string key)
        {
            string value = string.Empty;
            m_values.TryGetValue(key, out value);
            return value;
        }

        public override void SetString(string key, string value)
        {
            if (m_values.ContainsKey(key))
                m_values[key] = value;
            else
                m_values.Add(key, value);
        }

        public override void Apply()
        {
            Save();     
        }

        public override void Cancel()
        {
            m_values = new Dictionary<string, string>();
        }

        private void Load()
        {
            m_values = new Dictionary<string, string>();
            try
            {
                StreamReader stream = File.OpenText(m_childPath);
                string lineString;
                while ((lineString = stream.ReadLine()) != null)
                {
                    lineString = lineString.Trim();
                    string[] fields = lineString.Split('=');
                    if (fields.Length == 2 && lineString[0] != ';')
                    {
                        m_values.Add(fields[0], fields[1]);
                    }
                }
                stream.Close();
            }
            catch { }
        }

        private void Save()
        {
            if (m_values == null || m_values.Count == 0)
            {
                if (File.Exists(m_childPath))
                    File.Delete(m_childPath);
            }
            else
            {
                try
                {
                    StreamWriter stream = File.CreateText(m_childPath);
                    foreach (string key in m_values.Keys)
                    {
                        stream.WriteLine(string.Format("{0}={1}", key, m_values[key]));
                    }
                    stream.Close();

                }
                catch { }
            }
        }

    }
    #endregion

}
