﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Security.Principal;


// http://www.codeproject.com/KB/threads/csthreadmsg.aspx

namespace LockDownSvc
{
    /// <summary>
    /// Shared memory wrapper - creates a block of shared memory with basic API to access memory.
    /// 
    /// Author: Dennis Lang
    /// LockDown service - Dec 2010
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public class SharedMemory
    {
        /// <summary>
        ///   Specifies page protection for the mapped file
        ///   These correspond to the PAGE_XXX set of flags passed to CreateFileMapping()
        /// </summary>
        [Flags]
        public enum MapProtection
        {
            PageNone = 0x00000000,
            // protection
            PageReadOnly = 0x00000002,
            PageReadWrite = 0x00000004,
            PageWriteCopy = 0x00000008,
            // attributes
            SecImage = 0x01000000,
            SecReserve = 0x04000000,
            SecCommit = 0x08000000,
            SecNoCache = 0x10000000,
        }

        /// <summary>
        ///   Specifies access for the mapped file.
        ///   These correspond to the FILE_MAP_XXX constants used by MapViewOfFile[Ex]()
        /// </summary>
        public enum MapViewAccess
        {
            FileMapCopy = 0x0001,
            FileMapWrite = 0x0002,
            FileMapRead = 0x0004,
            FileMapAllAccess = 0x001f,
        }
  
        [DllImport("Kernel32.dll", EntryPoint = "CreateFileMapping",
         SetLastError = true, CharSet = CharSet.Unicode)]
        internal static extern IntPtr CreateFileMapping(uint hFile,
            // IntPtr lpAttributes, 
            ref SECURITY_ATTRIBUTES lpSecurityAttributes,
            uint flProtect,
          uint dwMaximumSizeHigh, uint dwMaximumSizeLow, string lpName);

        [DllImport("kernel32.dll", EntryPoint = "OpenFileMapping",
            SetLastError = true, CharSet = CharSet.Auto)]
        private static extern IntPtr OpenFileMapping(uint
            wDesiredAccess, bool bInheritHandle, String lpName);

        [DllImport("Kernel32.dll", EntryPoint = "MapViewOfFile",
         SetLastError = true, CharSet = CharSet.Unicode)]
        internal static extern IntPtr MapViewOfFile(IntPtr hFileMappingObject,
          uint dwDesiredAccess, uint dwFileOffsetHigh,
          uint dwFileOffsetLow, uint dwNumberOfBytesToMap);

        [DllImport("Kernel32.dll", EntryPoint = "UnmapViewOfFile",
         SetLastError = true, CharSet = CharSet.Unicode)]
        [return: MarshalAs(UnmanagedType.VariantBool)]
        internal static extern bool UnmapViewOfFile(IntPtr lpBaseAddress);

        [DllImport("kernel32.dll", EntryPoint = "CloseHandle",
            SetLastError = true, CharSet = CharSet.Auto)]
        private static extern bool CloseHandle(uint hHandle);
        
        [DllImport("kernel32.dll", EntryPoint = "GetLastError",
            SetLastError = true, CharSet = CharSet.Auto)]
        private static extern uint GetLastError();

#if false
        [DllImport("advapi32.dll", CharSet = CharSet.Auto, SetLastError = true,
            EntryPoint = "ConvertSecurityDescriptorToStringSecurityDescriptor",
            ExactSpelling = false)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool ConvertSecurityDescriptorToStringSecurityDescriptor(
            [In()] byte[] pSelfRelativeSD,
            int RequestedStringSDRevision,
            int SecurityInformation,
            out IntPtr StringSecurityDescriptor,
            out int StringSecurityDescriptorLen
        );

        public static string ConvertSDToStringSD(byte[] securityDescriptor, int securityInfo)
        {
            IntPtr pStringSD;
            int stringSDLen;
            bool success = ConvertSecurityDescriptorToStringSecurityDescriptor(
              securityDescriptor, 1, securityInfo, out pStringSD, out stringSDLen
              );
            // The following ensures that the memory allocated to pStringSD by the unmanaged 
            // Win32 API is freed. 
            try
            {
                if (!success)
                    throw new System.ComponentModel.Win32Exception(Marshal.GetLastWin32Error());
                return Marshal.PtrToStringAuto(pStringSD, stringSDLen);
            }
            finally
            {
                if (pStringSD != IntPtr.Zero)
                    Marshal.FreeHGlobal(pStringSD);
                pStringSD = IntPtr.Zero;
            }
        }
#endif

        [DllImport("advapi32.dll", CharSet = CharSet.Auto, SetLastError = true,
            EntryPoint = "ConvertStringSecurityDescriptorToSecurityDescriptor",
            ExactSpelling = false)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool ConvertStringSecurityDescriptorToSecurityDescriptor(
          [In(), MarshalAs(UnmanagedType.LPTStr)] string StringSecurityDescriptor,
          [In()] int StringSDRevision,
            // NOTE: The unmanaged API allocates the memory for this 
            // buffer and makes you responsible for cleaning it up. This 
            // is why the 'ref BYTE[]' did not work in your code below, 
            // because the CLR allocated a new ref BYTE[] on input, but 
            // the API over-wrote the address of the input pointer with 
            // the address of the memory that it allocated for the 
            // security descriptor buffer. 
          [Out()] out IntPtr SecurityDescriptor,
          [Out()] out int SecurityDescriptorSize
        );

        public static IntPtr /* byte[] */ ConvertStringSDToSD(string securityDescriptor)
        {
            IntPtr pSecurityDescriptor;
            int securityDescriptorLen;
            bool success = ConvertStringSecurityDescriptorToSecurityDescriptor(
                securityDescriptor, 1, out pSecurityDescriptor, out securityDescriptorLen);

            // The following ensures that the memory allocated to pSecurityDescriptor 
            // by the unmanaged Win32 API is freed. 
            try
            {
                if (!success)
                    throw new System.ComponentModel.Win32Exception(Marshal.GetLastWin32Error());

#if false
                byte[] sd = new byte[securityDescriptorLen];
                Marshal.Copy(pSecurityDescriptor, sd, 0, securityDescriptorLen);
                return sd;
#else
                return pSecurityDescriptor;
#endif
            }
            finally
            {
#if false
                if (pSecurityDescriptor != IntPtr.Zero)
                    Marshal.FreeHGlobal(pSecurityDescriptor);
                pSecurityDescriptor = IntPtr.Zero;
#endif
            }
        } 


        private IntPtr m_fileMapping;
        private IntPtr m_mappedView;

        /// <summary>
        /// Create a named shared memory region.
        /// </summary>
        /// <param name="name">ex: Global\\foo </param>
        /// <param name="size"></param>
        public SharedMemory(string name, uint size)
        {
            uint highSize = 0;
            uint lowSize = size;
            uint mapAccess = (uint)MapProtection.PageReadWrite;

            SECURITY_ATTRIBUTES sa = new SECURITY_ATTRIBUTES();
            sa.nLength = (uint)Marshal.SizeOf(sa);

            // Define the SDDL for the DACL. This example sets 
            // the following access:
            // Built-in guests are denied all access.
            // Anonymous logon is denied all access.
            // Authenticated users are allowed 
            // read/write/execute access.
            // Administrators are allowed full control.
            // Modify these values as needed to generate the proper
            // DACL for your application. 
            // "D:                  // Discretionary ACL
            // "(A;OICI;GA;;;WD)"   // Allow full control 
            string strSD = "D:(A;OICI;GA;;;WD)";

            sa.lpSecurityDescriptor = ConvertStringSDToSD(strSD);

            m_fileMapping = CreateFileMapping(0xFFFFFFFFu, ref sa, mapAccess, highSize, lowSize, name);
            if (m_fileMapping == IntPtr.Zero)
            {
                string errorMsg = String.Format("CFM Error Code: {0}", Marshal.GetLastWin32Error());
                m_fileMapping = OpenFileMapping((uint)MapViewAccess.FileMapWrite, false, name);
                if (m_fileMapping == IntPtr.Zero)
                {
                    errorMsg = String.Format("OFM Error Code: {0}", Marshal.GetLastWin32Error());
                }
                throw new FileLoadException(errorMsg);
            }

            if (sa.lpSecurityDescriptor != null)
            {
                Marshal.FreeHGlobal(sa.lpSecurityDescriptor);
                sa.lpSecurityDescriptor = IntPtr.Zero;
            }

            uint lowOffset = 0;
            uint highOffset = 0;
            uint viewAccess = (uint)MapViewAccess.FileMapWrite;   
            m_mappedView = MapViewOfFile(m_fileMapping, viewAccess, highOffset, lowOffset, (uint)size);
        }

        /// <summary>
        /// Return raw pointer to top of shared memory region.
        /// </summary>
        /// <returns></returns>
        public IntPtr GetPointer()
        {
            return m_mappedView;
        }

        public byte ReadByte(int offset)
        {
            return Marshal.ReadByte(m_mappedView, offset);
        }
        public void WriteByte(byte data, int offset)
        {
            Marshal.WriteByte(m_mappedView, offset, data);
        }

        public int ReadInt32(int offset)
        {
            return Marshal.ReadInt32(m_mappedView, offset);
        }
        public void WriteInt32(int data, int offset)
        {
            Marshal.WriteInt32(m_mappedView, offset, data);
        }

        public Int64 ReadInt64(int offset)
        {
            return Marshal.ReadInt64(m_mappedView, offset);
        }
        public void WriteInt64(Int64 data, int offset)
        {
            Marshal.WriteInt64(m_mappedView, offset, data);
        }

        public void ReadBytes(byte[] data, int offset)
        {
            for (int i = 0; i < data.Length; i++)
                data[i] = Marshal.ReadByte(m_mappedView, offset + i);
        }
        public void WriteBytes(byte[] data, int offset)
        {
            for (int i = 0; i < data.Length; i++)
                Marshal.WriteByte(m_mappedView, offset + i, data[i]);
        }

        public object ReadDeserialize(int offset, int length)
        {
            byte[] binaryData = new byte[length];
            ReadBytes(binaryData, offset);

            BinaryFormatter formatter = new BinaryFormatter();
            MemoryStream ms = new MemoryStream(binaryData, 0, length, true, true);
            object data = formatter.Deserialize(ms);
            ms.Close();
            return data;
        }

        public long WriteSerialize(object data, int offset, int length)
        {
            BinaryFormatter formatter = new BinaryFormatter();
            byte[] binaryData = new byte[length];
            MemoryStream ms = new MemoryStream(binaryData, 0, length, true, true);
            formatter.Serialize(ms, data);
            ms.Flush();
            ms.Close();
            WriteBytes(binaryData, offset);
            return ms.Length;
        }
    }
}

