namespace LockDownSvc
{
    using System;
    using System.Runtime.InteropServices;

    /// <summary>
    /// Writing a Parental Control Software for Windows
    /// By Pascal Ganaye | 8 Jul 2010
    /// Posted on CodeProject
    /// http://www.codeproject.com/
    /// 
    /// Used by: Dennis Lang
    /// LockDown service - Dec 2010
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public sealed class WinExitWindows
    {
        #region Public Methods
        static public bool ShutDown(ShutDownMethod Method, ShutDownReason Reason)
        {
            // Adjust the process privilege
            bool ret = AdjustShutdownTokenPriveleges();

            // Exit windows
            return ExitWindowsEx(Method, Reason);
        }

        static public bool LogOffInteractiveUser()
        {
            // Adjust the process privilege
            bool ret = AdjustShutdownTokenPriveleges();

            // Log off the interactive users.
            return ExitWindows(0, 0);
        }

        #endregion Public Methods

        #region Private Methods
        static private bool AdjustShutdownTokenPriveleges()
        {
            bool ret;

            IntPtr hProc = IntPtr.Zero;
            IntPtr hToken = IntPtr.Zero;
            LUID luidRestore;

            // Get the current current process security token
            hProc = System.Diagnostics.Process.GetCurrentProcess().Handle;
            ret = OpenProcessToken(hProc, TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, out hToken);
            if (!ret)
                return false;

            // Lookup the LUID for the shutdown privilege
            ret = LookupPrivilegeValue(String.Empty, SE_SHUTDOWN_NAME, out luidRestore);
            if (!ret)
            {
                Log.LogEvent("Shutdown lookup error " + Marshal.GetLastWin32Error());
                return false;
            }

            // Adjust the privileges of the current process to include the shutdown privilege.
            TOKEN_PRIVILEGES tokenPriviliges = new TOKEN_PRIVILEGES();
            tokenPriviliges.PrivilegeCount = 1;
            tokenPriviliges.Luid = luidRestore;
            tokenPriviliges.Attributes = SE_PRIVILEGE_ENABLED;

            return AdjustTokenPrivileges(hToken, false, ref tokenPriviliges, 0, IntPtr.Zero, IntPtr.Zero);
        }
        #endregion Private Methods

        #region ShutDown argument type definitions
        [Flags]
        public enum ShutDownMethod : uint
        {
            LogOff = 0x00,
            ShutDown = 0x01,
            Reboot = 0x02,
            Force = 0x04,
            PowerOff = 0x08,
            ForceIfHung = 0x10
        }

        [Flags]
        public enum ShutDownReason : uint
        {
            MajorApplication = 0x00040000,
            MajorHardware = 0x00010000,
            MajorLegacyApi = 0x00070000,
            MajorOperatingSystem = 0x00020000,
            MajorOther = 0x00000000,
            MajorPower = 0x00060000,
            MajorSoftware = 0x00030000,
            MajorSystem = 0x00050000,

            MinorBlueScreen = 0x0000000F,
            MinorCordUnplugged = 0x0000000b,
            MinorDisk = 0x00000007,
            MinorEnvironment = 0x0000000c,
            MinorHardwareDriver = 0x0000000d,
            MinorHotfix = 0x00000011,
            MinorHung = 0x00000005,
            MinorInstallation = 0x00000002,
            MinorMaintenance = 0x00000001,
            MinorMMC = 0x00000019,
            MinorNetworkConnectivity = 0x00000014,
            MinorNetworkCard = 0x00000009,
            MinorOther = 0x00000000,
            MinorOtherDriver = 0x0000000e,
            MinorPowerSupply = 0x0000000a,
            MinorProcessor = 0x00000008,
            MinorReconfig = 0x00000004,
            MinorSecurity = 0x00000013,
            MinorSecurityFix = 0x00000012,
            MinorSecurityFixUninstall = 0x00000018,
            MinorServicePack = 0x00000010,
            MinorServicePackUninstall = 0x00000016,
            MinorTermSrv = 0x00000020,
            MinorUnstable = 0x00000006,
            MinorUpgrade = 0x00000003,
            MinorWMI = 0x00000015,

            FlagUserDefined = 0x40000000,
            FlagPlanned = 0x80000000
        }
        #endregion ShutDown argument type definitions

        #region Win32 Imports
        [DllImport("user32.dll")]
        static extern bool ExitWindowsEx(ShutDownMethod uMethod, ShutDownReason dwReason);

        [DllImport("user32.dll")]
        static extern bool ExitWindows(Int32 uMethod, uint dwReason);

        [DllImport("advapi32.dll", SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        static extern bool AdjustTokenPrivileges(IntPtr TokenHandle,
            [MarshalAs(UnmanagedType.Bool)]bool DisableAllPrivileges,
            ref TOKEN_PRIVILEGES NewState,
            UInt32 BufferLength,
            IntPtr PreviousState,
            IntPtr ReturnLength);

        [DllImport("advapi32.dll", SetLastError = true)]
        static extern bool OpenProcessToken(IntPtr ProcessHandle,
            UInt32 DesiredAccess, out IntPtr TokenHandle);

        [DllImport("advapi32.dll", SetLastError = true, CharSet = CharSet.Auto)]
        static extern bool LookupPrivilegeValue(string lpSystemName, string lpName,
            out LUID lpLuid);

        [StructLayout(LayoutKind.Sequential)]
        private struct TOKEN_PRIVILEGES
        {
            public int PrivilegeCount;
            public LUID Luid;
            public int Attributes;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct LUID
        {
            public uint LowPart;
            public uint HighPart;
        }

        private const int TOKEN_QUERY = 0x08;
        private const int TOKEN_ADJUST_PRIVILEGES = 0x20;
        private const string SE_SHUTDOWN_NAME = "SeShutdownPrivilege";
        private const int SE_PRIVILEGE_ENABLED = 0x02;

        #endregion Win32 Imports
    }
}

