﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Security.Principal;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Management;

namespace LockDownSvc
{
    /// <summary>
    /// Collection of system level functions
    ///     GetUser()       ; return current user name else UnknownUser
    ///     UnknownUser     ; return UnknownUser string
    ///     GetUsers()      ; return string array of user accounts on local system
    ///     IsAdmin()       ; return true if current account is LockDown administrator
    /// 
    /// Author: Dennis Lang
    /// LockDown service - Dec 2010
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    class WinFunc
    {
        #region ==== GetUser functions

        /// <summary>
        /// Return unknown user string, used by GetUser()
        /// </summary>
        public static string UnknownUser
        {
            get { return "Unknown"; }
        }

        /// <summary>
        /// Get/Set foreground user name.
        /// </summary>
        public static string ForegroundUser
        {
            set 
            { 
                string[] userParts = value.Split('\\');
                s_ForegroundUser = userParts[userParts.Length - 1];
            }
            get { return (s_ForegroundUser.Length == 0) ? UnknownUser : s_ForegroundUser; }
        }
        private static string s_ForegroundUser = string.Empty;

        /// <summary>
        /// Return current interactive account user name, else unknown.
        /// </summary>
        /// <returns></returns>
        public static string GetUser()
        {
            string user = GetUser("explorer");
            if (user == UnknownUser)
                user = GetUser("winlogon");
            return user;
        }

        public static string GetUser(string processName)
        {
            // Returns user account of service, not neccessaryily the user's account WindowsIdentity.GetCurrent().Name;
            try
            {
                IntPtr foregroundWnd = GetForegroundWindow();
                int activeSessionId = WTSGetActiveConsoleSessionId();

                Process[] explorerProc = System.Diagnostics.Process.GetProcessesByName(processName);
                foreach (Process process in explorerProc)
                {
                    // if (process.MainWindowHandle != foregroundWnd)
                    if (process.SessionId != activeSessionId)
                        continue;

                    string user = string.Empty;

#if false
                    if (user == null || user.Length == 0)
                        user = process.StartInfo.UserName;
#endif
                    if (user == null || user.Length == 0)
                        user = GetProcessUser(process);

                    if (user == null || user.Length == 0)
                        user = process.StartInfo.EnvironmentVariables["username"];

                    if (user == null || user.Length == 0)
                        return UnknownUser;

                    return user;
                }
            }
            catch 
            { }

            return ForegroundUser;
        }

        /// <summary>
        /// Return user name associated with process.
        /// </summary>
        /// <param name="proc"></param>
        /// <returns></returns>
        public static string GetProcessUser(Process proc)
        {
            IntPtr AnswerBytes;
            IntPtr AnswerCount;
            if (WTSQuerySessionInformationW(WTS_CURRENT_SERVER_HANDLE,
                    proc.SessionId,
                    WTS_UserName,
                    out AnswerBytes,
                    out AnswerCount))
            {
                string userName = Marshal.PtrToStringUni(AnswerBytes);
                return userName;
            }
            else
            {
                return "unknown";
            }
        }

        #region ==== Window imports
        [DllImport("user32.dll")]
        public static extern IntPtr GetForegroundWindow();

        [DllImport("kernel32.dll")]
        static extern int WTSGetActiveConsoleSessionId();

        [DllImport("Wtsapi32.dll")]
        public static extern bool WTSQuerySessionInformationW(
            IntPtr hServer,
            int SessionId, 
            int WTSInfoClass,
            out IntPtr ppBuffer,
            out IntPtr pBytesReturned);

        // Just use the current TS server context.
        internal static IntPtr WTS_CURRENT_SERVER_HANDLE = IntPtr.Zero;

        // The User Name is the info we want returned by the query.
        internal static int WTS_UserName = 5;
        #endregion

        #endregion

        /// <summary>
        /// Return array of user accounts on local system.
        /// </summary>
        /// <returns></returns>
        public static string[] GetUsers()
        {
            lock (s_userNames)
            {
                if (s_userNames.Count == 0)
                {
                    s_userNames = new List<string>();
                    try
                    {
                        SelectQuery sQuery = new SelectQuery("Win32_UserAccount");
                        // SelectQuery sQuery = new SelectQuery("Win32_UserAccount", "LocalAccount=1");
                        // ManagementObjectSearcher mSearcher = new ManagementObjectSearcher(sQuery);
                        ManagementObjectSearcher mSearcher = new ManagementObjectSearcher(
                            "root\\CIMV2", "SELECT * FROM Win32_UserAccount where Disabled=\"False\" and LocalAccount=\"True\""); 

                        foreach (ManagementObject mObject in mSearcher.Get())
                        {
                            var varName = mObject["Name"];
                            string usrName = varName.ToString();
                            string LocalAccount = mObject["LocalAccount"].ToString();
                            if (!s_userNames.Contains(usrName))
                                s_userNames.Add(usrName);
                        }
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(ex.ToString());
                    }
                }

                return s_userNames.ToArray();
            }
        }
        static List<string> s_userNames = new List<string>();

#if false
        /// <summary>
        /// Check if current user has admin priviledges.
        /// NOTE - Does not work for services because it does not return interactive user.
        /// </summary>
        /// <returns></returns>
        public static bool IsAdmin2()
        {
            //bool value to hold our return value
            bool isAdmin;
            try
            {
                WindowsIdentity user = WindowsIdentity.GetCurrent();
                WindowsPrincipal principal = new WindowsPrincipal(user);
                isAdmin = principal.IsInRole(WindowsBuiltInRole.Administrator);
            }
            catch (UnauthorizedAccessException ex)
            {
                isAdmin = false;
            }
            catch (Exception ex)
            {
                isAdmin = false;
            }
            return isAdmin;
        }
#endif

    public static bool Is64bitPlatform { get { return (IntPtr.Size == 8 ? true : false); } }

    [StructLayout(LayoutKind.Sequential, Pack = 1)]
    public class SERVICE_STATUS
    {
      public uint dwServiceType = (uint)SERVICE_TYPE.SERVICE_WIN32_OWN_PROCESS;
      public uint dwCurrentState = (uint)SERVICE_STATE.SERVICE_RUNNING;
      public uint dwControlsAccepted = (uint)(
          SERVICE_CONTROLS_ACCEPTED.SERVICE_ACCEPT_STOP
        | SERVICE_CONTROLS_ACCEPTED.SERVICE_ACCEPT_POWEREVENT
        | SERVICE_CONTROLS_ACCEPTED.SERVICE_ACCEPT_SHUTDOWN
        //| SERVICE_CONTROLS_ACCEPTED.SERVICE_ACCEPT_PAUSE_CONTINUE
        //| SERVICE_CONTROLS_ACCEPTED.SERVICE_ACCEPT_SESSIONCHANGE
        );
      public uint dwWin32ExitCode = (uint)WIN32_ERROR.NO_ERROR;
      public uint dwServiceSpecificExitCode = 0;
      public uint dwCheckPoint = 0;
      public uint dwWaitHint;

      public SERVICE_STATUS() { }
    }

    [DllImport("advapi32.dll", SetLastError = true)]
    public static extern bool SetServiceStatus(
      IntPtr hServiceStatus,  // SERVICE_STATUS_HANDLE
      SERVICE_STATUS lpServiceStatus
      );

    public enum SERVICE_STATE : uint
    {
      SERVICE_STOPPED = 0x00000001,
      SERVICE_START_PENDING = 0x00000002,
      SERVICE_STOP_PENDING = 0x00000003,
      SERVICE_RUNNING = 0x00000004,
      SERVICE_CONTINUE_PENDING = 0x00000005,
      SERVICE_PAUSE_PENDING = 0x00000006,
      SERVICE_PAUSED = 0x00000007,
    }

    public enum WIN32_ERROR : uint
    {
      NO_ERROR = 0,
      ERROR_CALL_NOT_IMPLEMENTED = 120,
      ERROR_INSUFFICIENT_BUFFER = 122,
      ERROR_NO_MORE_ITEMS = 259, // 0x103
      ERROR_INVALID_DATA = 13,
      ERROR_SERVICE_SPECIFIC_ERROR = 1066,
    }

    public enum ACCESS_TYPE : uint
    {
      DELETE = (0x00010000),
      READ_CONTROL = (0x00020000),
      WRITE_DAC = (0x00040000),
      WRITE_OWNER = (0x00080000),
      SYNCHRONIZE = (0x00100000),

      STANDARD_RIGHTS_REQUIRED = (0x000F0000),

      STANDARD_RIGHTS_READ = (READ_CONTROL),
      STANDARD_RIGHTS_WRITE = (READ_CONTROL),
      STANDARD_RIGHTS_EXECUTE = (READ_CONTROL),

      STANDARD_RIGHTS_ALL = (0x001F0000),

      SPECIFIC_RIGHTS_ALL = (0x0000FFFF),

      // AccessSystemAcl access type
      ACCESS_SYSTEM_SECURITY = (0x01000000),

      // MaximumAllowed access type
      MAXIMUM_ALLOWED = (0x02000000),

      //  These are the generic rights.
      GENERIC_READ = (0x80000000),
      GENERIC_WRITE = (0x40000000),
      GENERIC_EXECUTE = (0x20000000),
      GENERIC_ALL = (0x10000000),
    }

    public enum SCM_ACCESS_TYPE : uint
    {
      SC_MANAGER_CONNECT = 0x0001,
      SC_MANAGER_CREATE_SERVICE = 0x0002,
      SC_MANAGER_ENUMERATE_SERVICE = 0x0004,
      SC_MANAGER_LOCK = 0x0008,
      SC_MANAGER_QUERY_LOCK_STATUS = 0x0010,
      SC_MANAGER_MODIFY_BOOT_CONFIG = 0x0020,
    }

    public enum SERVICE_ACCESS_TYPE : uint
    {
      SERVICE_QUERY_CONFIG = 0x0001,
      SERVICE_CHANGE_CONFIG = 0x0002,
      SERVICE_QUERY_STATUS = 0x0004,
      SERVICE_ENUMERATE_DEPENDENTS = 0x0008,
      SERVICE_START = 0x0010,
      SERVICE_STOP = 0x0020,
      SERVICE_PAUSE_CONTINUE = 0x0040,
      SERVICE_INTERROGATE = 0x0080,
      SERVICE_USER_DEFINED_CONTROL = 0x0100,
    }

    public enum STANDARD_ACCESS_TYPE : uint
    {
      DELETE = (0x00010000),
      READ_CONTROL = (0x00020000),
      WRITE_DAC = (0x00040000),
      WRITE_OWNER = (0x00080000),
      SYNCHRONIZE = (0x00100000),

      STANDARD_RIGHTS_REQUIRED = (0x000F0000),
    }

    public enum SERVICE_TYPE : uint
    {
      SERVICE_FILE_SYSTEM_DRIVER = 0x00000002, //The service is a file system driver. 
      SERVICE_KERNEL_DRIVER = 0x00000001, //The service is a device driver. 
      SERVICE_WIN32_OWN_PROCESS = 0x00000010, //The service runs in its own process. 
      SERVICE_WIN32_SHARE_PROCESS = 0x00000020, //The service shares a process with other services. 
      //If the service type is either SERVICE_WIN32_OWN_PROCESS or SERVICE_WIN32_SHARE_PROCESS, and the service is running in the context of the LocalSystem account, the following type may also be specified.
      SERVICE_INTERACTIVE_PROCESS = 0x00000100,
    }

    public enum SERVICE_CONTROL_CODES : uint
    {
      SERVICE_CONTROL_STOP = 0x00000001,
      SERVICE_CONTROL_PAUSE = 0x00000002,
      SERVICE_CONTROL_CONTINUE = 0x00000003,
      SERVICE_CONTROL_INTERROGATE = 0x00000004,
      SERVICE_CONTROL_SHUTDOWN = 0x00000005,
      SERVICE_CONTROL_PARAMCHANGE = 0x00000006,
      SERVICE_CONTROL_NETBINDADD = 0x00000007,
      SERVICE_CONTROL_NETBINDREMOVE = 0x00000008,
      SERVICE_CONTROL_NETBINDENABLE = 0x00000009,
      SERVICE_CONTROL_NETBINDDISABLE = 0x0000000A,
      SERVICE_CONTROL_DEVICEEVENT = 0x0000000B,
      SERVICE_CONTROL_HARDWAREPROFILECHANGE = 0x0000000C,
      SERVICE_CONTROL_POWEREVENT = 0x0000000D,
      SERVICE_CONTROL_SESSIONCHANGE = 0x0000000E,
    }

    public enum SERVICE_CONTROLS_ACCEPTED : uint
    {
      SERVICE_ACCEPT_NETBINDCHANGE = 0x00000010, /* The service is a network component that can accept changes in its binding without being stopped and restarted. 
This control code allows the service to receive SERVICE_CONTROL_NETBINDADD, SERVICE_CONTROL_NETBINDREMOVE, SERVICE_CONTROL_NETBINDENABLE, and SERVICE_CONTROL_NETBINDDISABLE notifications.

Windows NT:  This value is not supported. */
      SERVICE_ACCEPT_PARAMCHANGE = 0x00000008, /* The service can reread its startup parameters without being stopped and restarted. 
This control code allows the service to receive SERVICE_CONTROL_PARAMCHANGE notifications.

Windows NT:  This value is not supported. */
      SERVICE_ACCEPT_PAUSE_CONTINUE = 0x00000002, /* The service can be paused and continued. 
This control code allows the service to receive SERVICE_CONTROL_PAUSE and SERVICE_CONTROL_CONTINUE notifications. */

      SERVICE_ACCEPT_SHUTDOWN = 0x00000004, /* The service is notified when system shutdown occurs. 
This control code allows the service to receive SERVICE_CONTROL_SHUTDOWN notifications. Note that ControlService cannot send this notification; only the system can send it. */

      SERVICE_ACCEPT_STOP = 0x00000001, /* The service can be stopped. 
This control code allows the service to receive SERVICE_CONTROL_STOP notifications. */

      //This value can also contain the following extended control codes, which are supported only by HandlerEx.

      SERVICE_ACCEPT_HARDWAREPROFILECHANGE = 0x00000020, /* The service is notified when the computer's hardware profile has changed. This enables the system to send SERVICE_CONTROL_HARDWAREPROFILECHANGE notifications to the service.
Windows NT:  This value is not supported. */
      SERVICE_ACCEPT_POWEREVENT = 0x00000040, /* The service is notified when the computer's power status has changed. This enables the system to send SERVICE_CONTROL_POWEREVENT notifications to the service.
Windows NT:  This value is not supported. */
      SERVICE_ACCEPT_SESSIONCHANGE = 0x00000080, /* The service is notified when the computer's session status has changed. This enables the system to send SERVICE_CONTROL_SESSIONCHANGE notifications to the service. */
    }

    public enum SERVICE_START_TYPE : uint
    {
      SERVICE_BOOT_START = 0x00000000,
      SERVICE_SYSTEM_START = 0x00000001,
      SERVICE_AUTO_START = 0x00000002,
      SERVICE_DEMAND_START = 0x00000003,
      SERVICE_DISABLED = 0x00000004,
    }

    public enum SERVICE_ERROR_CODE : uint
    {
      SERVICE_ERROR_IGNORE = 0x00000000,
      SERVICE_ERROR_NORMAL = 0x00000001,
      SERVICE_ERROR_SEVERE = 0x00000002,
      SERVICE_ERROR_CRITICAL = 0x00000003,
    }

    public static uint SERVICE_ALL_ACCESS = ((uint)STANDARD_ACCESS_TYPE.STANDARD_RIGHTS_REQUIRED |
      (uint)WinFunc.SERVICE_ACCESS_TYPE.SERVICE_QUERY_CONFIG |
      (uint)WinFunc.SERVICE_ACCESS_TYPE.SERVICE_CHANGE_CONFIG |
      (uint)WinFunc.SERVICE_ACCESS_TYPE.SERVICE_QUERY_STATUS |
      (uint)WinFunc.SERVICE_ACCESS_TYPE.SERVICE_ENUMERATE_DEPENDENTS |
      (uint)WinFunc.SERVICE_ACCESS_TYPE.SERVICE_START |
      (uint)WinFunc.SERVICE_ACCESS_TYPE.SERVICE_STOP |
      (uint)WinFunc.SERVICE_ACCESS_TYPE.SERVICE_PAUSE_CONTINUE |
      (uint)WinFunc.SERVICE_ACCESS_TYPE.SERVICE_INTERROGATE |
      (uint)WinFunc.SERVICE_ACCESS_TYPE.SERVICE_USER_DEFINED_CONTROL);


        /// ----------------------------
        [ DllImport("PowrProf")]
        extern static bool SetSuspendState(bool hibernate, bool force , bool disablewakeup);
        /*
        SetSuspendState(true, true, false); // force hibernate, disable all wakeup events
        SetSuspendState(false, true, true); // force suspend and enable wake events.
        SetSuspendState(false, false, true); // broadcast PBT_APMQUERYSUSPEND event and suspend if all agree.
        */


    }
}
