﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Media;
using Microsoft.Win32;
using System.IO;
using System.Collections;
using System.Drawing.Imaging;

namespace LockDownSvc
{
    /// <summary>
    /// Simple Alarm Clock.
    /// 
    /// Author: Dennis Lang
    /// LockDown service - Dec 2010
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public partial class AlarmDlg : Form
    {
        private DateTime m_alarmAt = DateTime.MinValue;
        private DateTime m_startAt = DateTime.Now;
        private RegMgr m_regMgr;
        private DataMgr m_regSchdMgr;
       
        /// Constant strings used to store in registry and to populate buttons and stuff.
        const string s_alarm = @"Software\LockDown\Alarm";
        const string s_font = "Font";
        const string s_yes = "Yes";
        const string s_no = "No";
        const string s_dur = "Dur";
        const string s_at = "At";
        const string s_ElapsedTitle = "Elapsed";
        const string s_LeftTitle = "Left";
        const int s_subFieldCnt = 10;

        string[] timeFmts = new string[] { "H:mm:ss", "hh:mm:ss tt" };
        int m_timeFmt = 0;
        int m_alarmFmt = 0;

        public class OptionInfo
        {
            // ---- Image options ----
            public ImageLayout imageLayout = ImageLayout.Zoom;  // none, tile, center, stretch, zoom
            public Point imagePosition = Point.Empty;           // only used if layout is none
            public TimeSpan imageMotionTs = TimeSpan.Zero;
            public PointF imageMotionDir = PointF.Empty;        // seconds, minutes, hours, days
            public bool isFillMargin = true;
            public Point imageFillPixel = new Point(-1, -1);
            public Color imageFillColor = Color.Black;

            // ---- Text options ----
            public enum TextMode { eText, eTimeLeft, eTimeAt };
            public TextMode textMode = TextMode.eText;
            public const int s_centerTextPointX = -1;
            public const int s_centerTextPointY = -1;
            public Point textPosition = new Point(OptionInfo.s_centerTextPointX, OptionInfo.s_centerTextPointY);
        }

        OptionInfo m_optionInfo = new OptionInfo();

        #region ==== Public interface

        public AlarmDlg()
        {
            InitializeComponent();

            elapsedLeftTitle.Text = s_LeftTitle;

            this.scheduleView.ListViewItemSorter = new ListViewItemComparer(0);
            this.scheduleView.Sorting = SortOrder.Ascending;

            soundTbox.Text = @"C:\Windows\Media\chimes.wav";

            durationGrp.Enabled = durationBtn.Checked = true;
            atGrp.Enabled = atTimeBtn.Checked = false;
           

            m_regMgr = new RegMgr(s_alarm, Registry.CurrentUser);
            m_regSchdMgr = m_regMgr.Clone().SetChild("Schedule");
            LoadSchedule();
            LoadOptions();

            if (scheduleView.Items.Count == 0)
                AddSchdItem_Click(null, EventArgs.Empty);
            else
                scheduleView.Items[0].Selected = true;

            AlarmEnabled = true;
            // Parse();
            timer_Tick(null, EventArgs.Empty);
        }

        /// <summary>
        /// Fire when alarm triggers.
        /// </summary>
        public EventHandler AlarmTriggered;

        /// <summary>
        /// Fire when alaram enable state changes.
        /// </summary>
        public EventHandler AlarmStateChanged;     // on/off

        /// <summary>
        /// Set/Get Alarm enable state.
        /// </summary>
        public bool AlarmEnabled
        {
            get { return alarmAtCk.Checked && scheduleView.CheckedItems.Count != 0; }
            set 
            {
                if (alarmAtCk.Checked != value)
                {
                    alarmAtCk.Checked = value;
                    TriggerStateChange();
                }
            }
        }
        #endregion

        #region ==== Buttons

        private void durationBtn_CheckedChanged(object sender, EventArgs e)
        {
            durationGrp.Enabled = durationBtn.Checked;
        }

        private void atTimeBtn_CheckedChanged(object sender, EventArgs e)
        {
            atGrp.Enabled = atTimeBtn.Checked;
            Parse();
        }

        private void closeBtn_Click(object sender, EventArgs e)
        {
            this.Close();
        }
        #endregion

        #region ==== Close and Timer
        /// <summary>
        /// Prevent close from disposing dialog. Just hide dialog.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnClosing(CancelEventArgs e)
        {
            SaveSchedule();
            SaveOptions();

            e.Cancel = true;
            this.Visible = false;
            base.OnClosing(e);
        }


        public DateTime AlarmAt { get; private set; }

        private void timer_Tick(object sender, EventArgs e)
        {
            DateTime now = DateTime.Now;
            timeLbl.Text = now.ToString(timeFmts[m_timeFmt]);

            // Sort shceduleView
            // set ui to top of list (closest to triggering)
            scheduleView.Sort();

            foreach (ListViewItem lvItem in scheduleView.Items)
            {
                int schdIdx = lvItem.Index;
                DateTime alarmAt = (DateTime)lvItem.Tag;

                if (lvItem.Selected)
                {
                    if (elapsedLeftTitle.Text == s_ElapsedTitle)
                    {
                        TimeSpan elapsed = now - m_startAt;
                        elapsedLbl.Text = string.Format("{0}:{1:00}:{2:00}",
                            elapsed.Hours, elapsed.Minutes, elapsed.Seconds);
                        this.Text = elapsedLbl.Text;
                    }
                    else
                    {
                        TimeSpan left = alarmAt - now;
                        if (left.TotalSeconds >= 0)
                        {
                            if (left.TotalDays > 1)
                                elapsedLbl.Text = string.Format("{0:N2} days", left.TotalDays);
                            else
                                elapsedLbl.Text = string.Format("{0}:{1:00}:{2:00}",
                                    left.Hours, left.Minutes, left.Seconds);
                            this.Text = elapsedLbl.Text;
                        }
                        else
                            this.Text = elapsedLbl.Text = "Alarm!";
                    }
                }

                if (now >= alarmAt)
                {
                    // LoadFromSchedule(schdIdx);
                    bool repeat = (lvItem.SubItems[colRepeat.Index].Text == s_yes);

                    if (AlarmEnabled && lvItem.Checked)
                    {
                        string soundStr = lvItem.SubItems[colSnd.Index].Text;
                        string imageStr = lvItem.SubItems[colImage.Index].Text;
                        string textStr = lvItem.SubItems[colText.Index].Text;
                        string executeStr = lvItem.SubItems[colExe.Index].Text;

                        SoundAlarm(soundStr, true);
                        ShowAlertImage(imageStr, textStr);
                        ExecuteJob(executeStr);

                        // AlarmEnabled = !autoOffCk.Checked;
                        elapsedLbl.BackColor = Color.Red;

                        if (AlarmTriggered != null)
                            AlarmTriggered(this, EventArgs.Empty);
                    }

                    if (repeat)
                    {
                        // Repeat
                        if (lvItem.SubItems[colType.Index].Text == s_dur)
                        {
                            // Repeat duration
                            try
                            {
                                double durVal = double.Parse(lvItem.SubItems[colDur.Index].Text);
                                int durUnit = int.Parse(lvItem.SubItems[colDurUnit.Index].Text);
                                alarmAt = DateTime.Now + Duration(durVal, durUnit);
                            }
                            catch
                            {
                                alarmAt = DateTime.Now + TimeSpan.FromHours(1);
                                lvItem.SubItems[colDur.Index].Text = "1";
                                lvItem.SubItems[colDurUnit.Index].Text = "2";
                            }
                        }
                        else
                        {
                            // Repeat At (next day)
                            alarmAt += TimeSpan.FromDays(1);
                        }
                        lvItem.Tag = alarmAt;
                        string alarmStr = alarmAt.ToString("dd-MMM-yy HH:mm:ss tt");
                        lvItem.SubItems[colTime.Index].Text = alarmStr;
                    }
                    else
                    {
                        scheduleView.Items.Remove(lvItem);
                    }

                    TriggerStateChange();
                }
                else
                {
                    AlarmAt = alarmAt;
                    break;
                }
            }

            UpdateSchedule();
        }

        /// <summary>
        /// Toggle whether Elapsed or time Left is displayed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void elapsedLeftBtn_Click(object sender, EventArgs e)
        {
            if (elapsedLeftTitle.Text == s_ElapsedTitle)
            {
                elapsedLbl.ForeColor = elapsedLeftTitle.ForeColor = Color.DarkRed;
                elapsedLeftTitle.Text = s_LeftTitle;
            }
            else
            {
                elapsedLbl.ForeColor = elapsedLeftTitle.ForeColor = Color.DarkGreen;
                elapsedLeftTitle.Text = s_ElapsedTitle;
            }

            timer_Tick(null, EventArgs.Empty);
        }

      
        #endregion

        #region ==== Duration and TimeAt UI
        private void quickPickCompo_SelectedIndexChanged(object sender, EventArgs e)
        {
            string pick = quickPickCompo.SelectedItem.ToString();
            TimeSpan timeSpan = TimeSpan.Zero;
            uint hour = 0;

            switch (pick)
            {
                case "In 5 minutes":
                    timeSpan = new TimeSpan(0, 5, 0);
                    break;
                case "In 10 minutes":
                    timeSpan = new TimeSpan(0, 10, 0);
                    break;
                case "In 30 minutes":
                    timeSpan = new TimeSpan(0, 30, 0);
                    break;
                case "In 1 hour":
                    timeSpan = new TimeSpan(1, 0, 0);
                    break;
                case "In 2 hours":
                    timeSpan = new TimeSpan(2, 0, 0);
                    break;

                case "At 8am":
                    hour = 8;
                    break;
                case "At 10 am":
                    hour = 10;
                    break;
                case "At Noon":
                    hour = 12;
                    break;
                case "At 2 pm":
                    hour = 12 + 2;
                    break;
                case "At 4 pm":
                    hour = 12 + 4;
                    break;
                case "At 6 pm":
                    hour = 12 + 6;
                    break;
                case "At 8 pm":
                    hour = 12 + 8;
                    break;          
                case "At 10 pm":
                    hour = 12 + 10;
                    break;          
                case "At Midnight":
                    hour = 0;
                    break;

                default:
                    break;
            }

            if (timeSpan != TimeSpan.Zero)
            {
                durationBtn.Checked = true;
                if (timeSpan.TotalHours >= 1)
                {
                    durUnits.SelectedIndex = 2;
                    durTbox.Text = timeSpan.TotalHours.ToString("N2");
                }
                else if (timeSpan.TotalMinutes >= 1)
                {
                    durUnits.SelectedIndex = 1;
                    durTbox.Text = timeSpan.TotalMinutes.ToString("N2");
                }
                else
                {
                    durUnits.SelectedIndex = 0;
                    durTbox.Text = timeSpan.TotalMinutes.ToString("N2");
                }
            }
            else
            {
                hourNum.Value = hour;
                minNum.Value = 0;
            }
            m_startAt = DateTime.Now;
        }

        private bool m_skipParse = false;

        /// <summary>
        /// Convert value and units to a TimeSpan
        /// </summary>
        /// <param name="val"></param>
        /// <param name="units_smh">0=seconds,1=minutes,2=hours</param>
        /// <returns></returns>
        TimeSpan Duration(double val, int units_smh)
        {
            TimeSpan timeSpan = TimeSpan.Zero;

            switch (units_smh)
            {
                case 0: // seconds
                    timeSpan = TimeSpan.FromSeconds(val);
                    break;
                case 1: // minutes
                    timeSpan = TimeSpan.FromMinutes(val);
                    break;
                case 2: // hours
                    timeSpan = TimeSpan.FromHours(val);
                    break;
            }

            return timeSpan;
        }

        private void Parse()
        {
            if (m_skipParse)
                return;

            if (durationBtn.Checked)
            {
                try
                {
                    double val = string.IsNullOrEmpty(durTbox.Text) ? 0 : double.Parse(durTbox.Text);
                    switch (durUnits.SelectedIndex)
                    {
                        case 0: // seconds
                            m_alarmAt = DateTime.Now + TimeSpan.FromSeconds(val);
                            durTrackBar.Value = (val < 60) ? (int)val * 100 / 60 : 0;
                            break;
                        case 1: // minutes
                            m_alarmAt = DateTime.Now + TimeSpan.FromMinutes(val);
                            durTrackBar.Value = (val < 60) ? (int)val * 100 / 60 : 0;
                            break;
                        case 2: // hours
                            m_alarmAt = DateTime.Now + TimeSpan.FromHours(val);
                            durTrackBar.Value = (val < 24) ? (int)val * 100 / 24 : 0;
                            break;
                    }
                }
                catch
                {
                    durTbox.Text = "0";
                }
            }
            else
            {
                DateTime now = DateTime.Now;
                m_alarmAt = new DateTime(now.Year, now.Month, now.Day, (int)hourNum.Value, (int)minNum.Value, 0);
                if (m_alarmAt < now)
                    m_alarmAt += TimeSpan.FromDays(1);
            }

            this.m_startAt = DateTime.Now;
            alarmAtLbl.Text = this.m_alarmAt.ToString(timeFmts[m_alarmFmt]);
            elapsedLbl.BackColor = alarmAtLbl.BackColor;
            // AlarmEnabled = true;

            if (scheduleView.SelectedIndices.Count != 0)
                SetSchedule(scheduleView.SelectedIndices[0]);
        }

        private void durTbox_TextChanged(object sender, EventArgs e)
        {
            Parse();
        }

        private void durUnits_SelectedIndexChanged(object sender, EventArgs e)
        {
            Parse();
        }

        private void hourNum_ValueChanged(object sender, EventArgs e)
        {
            Parse();
        }

        private void minNum_ValueChanged(object sender, EventArgs e)
        {
            Parse();
        }

        private void trackBar_Scroll(object sender, EventArgs e)
        {
            int scale = (durUnits.SelectedIndex != 2) ? 60 : 24;
            durTbox.Text = (durTrackBar.Value * scale / 100).ToString("N0");
            Parse();
        }
        #endregion

        #region ==== Sound
        private void soundTbox_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            if (e.KeyCode == Keys.Tab)
                e.IsInputKey = true;
            else if (e.KeyCode == Keys.Enter)
                Parse();
        }

        private void soundBtn_Click(object sender, EventArgs e)
        {
            // this.openFileDialog.InitialDirectory = @"C:\Windows\Media";
            this.openFileDialog.Title = "Select Wav Sound File";
            this.openFileDialog.Filter = "wav|*.wav";
            this.openFileDialog.InitialDirectory = string.Empty;

            try
            {
                openFileDialog.FileName = soundTbox.Text;
                if (string.IsNullOrEmpty(openFileDialog.FileName))
                    this.openFileDialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyMusic);
                else 
                    this.openFileDialog.InitialDirectory = Path.GetDirectoryName(openFileDialog.FileName);
            }
            catch 
            {  }

            if (this.openFileDialog.InitialDirectory.Length == 0)
            {
                //  @"C:\Windows\Media";
                // Path.Combine(Path.GetDirectoryName(Application.ExecutablePath), "AlarmSounds");
                this.openFileDialog.InitialDirectory = Path.GetDirectoryName(Application.ExecutablePath);
            }

            if (openFileDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                soundTbox.Text = openFileDialog.FileName;
                Parse();
            }
        }

        public void SoundAlarm(string soundWav, bool loop)
        {
            try
            {
                if (soundWav.Trim().Length != 0)
                {
                    m_soundPlayer.SoundLocation = soundWav;
                    if (loop)
                        m_soundPlayer.Play();       // PlayLooping()
                    else
                        m_soundPlayer.Play();
                }
            }
            catch { }
        }
        SoundPlayer m_soundPlayer = new SoundPlayer();

        private void playBtn_Click(object sender, EventArgs e)
        {
            SoundAlarm(soundTbox.Text, false);
        }
        #endregion

        #region ==== Image
        private void imageBtn_Click(object sender, EventArgs e)
        {
            this.openFileDialog.Title = "Select Image File";
            this.openFileDialog.Filter = "any|*.*|png|*.png|jpg|*.jpg|bmp|*.bmp|gif|*.gif";
            this.openFileDialog.InitialDirectory = string.Empty;

            try
            {
                openFileDialog.FileName = imageTbox.Text;
                if (string.IsNullOrEmpty(openFileDialog.FileName))
                    this.openFileDialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyPictures);
                else
                    this.openFileDialog.InitialDirectory = Path.GetDirectoryName(openFileDialog.FileName);
 
            }
            catch { }

            if (this.openFileDialog.InitialDirectory.Length == 0)
            {
                // System.Environment.GetFolderPath(Environment.SpecialFolder.MyPictures);
                // Path.Combine(Path.GetDirectoryName(Application.ExecutablePath), "AlarmImages");
                this.openFileDialog.InitialDirectory = Path.GetDirectoryName(Application.ExecutablePath);
            }

            if (openFileDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                imageTbox.Text = openFileDialog.FileName;
                Parse();
            }
        }

        private void showImgBtn_Click(object sender, EventArgs e)
        {
            ShowAlertImage(imageTbox.Text, alarmTbox.Text);
            SoundAlarm(soundTbox.Text, true);
        }

        Form m_alertForm = null;
        PictureBox m_alertPbox = null;
        FrameDimension m_imageDimensions = null;
        int m_frameCnt = 0;
        int m_frameIdx = 0;

        private void AnimateNextFrame()
        {
            if (m_frameCnt != 0 && m_alertForm != null)
            {
                m_frameIdx = (m_frameIdx + 1) % m_frameCnt;
                m_alertPbox.BackgroundImage.SelectActiveFrame(m_imageDimensions, m_frameIdx);
                // m_alertPbox.Refresh();
                m_alertForm.Refresh();
            }
        }

        private void ShowAlertImage(string imagePath, string textMsg)
        {
            Size screenSz = System.Windows.Forms.Screen.PrimaryScreen.Bounds.Size;
                
            m_alertForm = new Form();
            m_alertForm.WindowState = FormWindowState.Maximized;
            m_alertForm.FormBorderStyle = FormBorderStyle.None;
            m_alertForm.TopMost = true;
            m_alertForm.BackColor = m_optionInfo.isFillMargin ? m_optionInfo.imageFillColor :  Color.LimeGreen;     // optional transparent
            m_alertForm.TransparencyKey = m_optionInfo.isFillMargin ? Color.Empty : m_alertForm.BackColor;
            m_alertForm.Opacity = 1.0;
            m_alertForm.AllowTransparency = !m_optionInfo.isFillMargin;

            m_alertPbox = new PictureBox();
            m_alertPbox.Dock = System.Windows.Forms.DockStyle.Fill;
            m_alertPbox.Location = m_optionInfo.imagePosition;
            m_alertPbox.Click += new EventHandler(Alert_Click);

            m_alertForm.Controls.Add(m_alertPbox);

            m_frameIdx = 0;
            try
            {
                m_alertPbox.BackgroundImageLayout = m_optionInfo.imageLayout;
                m_alertPbox.BackgroundImage = Bitmap.FromFile(imagePath);
                // Bitmap img = new Bitmap(m_alertPbox.BackgroundImage);
                // m_alertForm.BackColor = img.GetPixel(0, 0);      // where to get background color from
                m_imageDimensions = new FrameDimension(m_alertPbox.BackgroundImage.FrameDimensionsList[0]);
                m_frameCnt = m_alertPbox.BackgroundImage.GetFrameCount(m_imageDimensions);
                m_frameIdx = 0;
            }
            catch (Exception ex)
            {
                m_alertPbox.BackgroundImageLayout = ImageLayout.Stretch;
                m_alertPbox.Image = new Bitmap(screenSz.Width, screenSz.Height);
                Graphics gBg = Graphics.FromImage(m_alertPbox.Image);
                // gBg.Clear(Color.FromArgb(128, Color.Red));
                gBg.Clear(m_optionInfo.isFillMargin ? m_optionInfo.imageFillColor : Color.Transparent);
                gBg.Dispose();
                m_frameCnt = 0;
            }

            if (textMsg.Trim().Length != 0)
            {
                // Font font = new Font("Tahoma", 100);
                Font font = fontDialog.Font;
                StringFormat drawFormat = new StringFormat();
                drawFormat.Alignment = (m_optionInfo.textPosition.Y == OptionInfo.s_centerTextPointY) ?
                    StringAlignment.Center : StringAlignment.Near;
                drawFormat.LineAlignment = (m_optionInfo.textPosition.X == OptionInfo.s_centerTextPointX) ?
                    StringAlignment.Center : StringAlignment.Near;
                drawFormat.Trimming = StringTrimming.Character;

                m_alertPbox.Image = new Bitmap(screenSz.Width, screenSz.Height);
                m_alertPbox.SizeMode = PictureBoxSizeMode.CenterImage;
                Graphics gText = Graphics.FromImage(m_alertPbox.Image);
                gText.Clear(Color.Transparent);

                Rectangle rect = new Rectangle(m_optionInfo.textPosition, m_alertPbox.Image.Size);
                gText.DrawString(textMsg, font, Brushes.White, rect, drawFormat);
                gText.Dispose();
            }

            //  m_alertImage.MouseMove += new MouseEventHandler(m_alertImage_MouseMove);
            m_alertImageTm = DateTime.Now;
            m_alertForm.Show();
            m_alertForm.BringToFront();
            animateTimer.Enabled = true;    // (m_frameCnt != 0);
        }
       
        /// <summary>
        /// Alert image has been clicked, hide image and reset alarm.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Alert_Click(Object sender, EventArgs e)
        {
            if ((DateTime.Now - m_alertImageTm).TotalSeconds > 5)
            {
                m_alertForm.Close();
                m_alertForm = null;
                m_soundPlayer.Stop();
                animateTimer.Stop();
            }
#if false
            AlarmEnabled = false;
            if (repeatCk.Checked)
            {
                Parse(durationBtn.Checked ? ParseType.eDuration : ParseType.eAtTime);
            }
#endif
        }
        DateTime m_alertImageTm;
        #endregion


        private void ExecuteJob(string exePath)
        {
            try
            {
                if (File.Exists(exePath))
                {
                    System.Diagnostics.Process.Start(executeTbox.Text);
                }
            } catch {}
        }

        #region ==== Misc Font, CheckChanged
        /// <summary>
        /// Show font selection dialog.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void fontBtn_Click(object sender, EventArgs e)
        {
            fontDialog.ShowDialog();
        }

        private void alarmAtCk_CheckedChanged(object sender, EventArgs e)
        {
            alarmAtCk.ForeColor = (alarmAtCk.Checked ? Color.DarkGreen : Color.DarkRed);
            TriggerStateChange();
        }
        #endregion

        private void exeBtn_Click(object sender, EventArgs e)
        {
            this.openFileDialog.Title = "Select Executable";
            this.openFileDialog.Filter = "exe|*.exe|any|*.*";
            this.openFileDialog.InitialDirectory = string.Empty;

            try
            {
                openFileDialog.FileName = executeTbox.Text;
                this.openFileDialog.InitialDirectory = Path.GetDirectoryName(openFileDialog.FileName);
            }
            catch
            { }

            if (this.openFileDialog.InitialDirectory.Length == 0)
                this.openFileDialog.InitialDirectory = System.Environment.GetFolderPath(Environment.SpecialFolder.MyComputer);

            if (openFileDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                executeTbox.Text = openFileDialog.FileName;
                Parse();
            }
        }

        private void runExeBtn_Click(object sender, EventArgs e)
        {
            ExecuteJob(executeTbox.Text);
        }

      
        private string GetFilename(string path)
        {
            try
            {
                // return Path.GetFileName(path);
                return path;
            }
            catch { }
            return string.Empty;
        }

        private void UpdateSchedule()
        {
            foreach (ListViewItem lvItem in scheduleView.Items)
            {
                DateTime alarmAt = (DateTime)lvItem.Tag;
                // TimeSpan elapsed = DateTime.Now - m_startAt;
                TimeSpan left = alarmAt - DateTime.Now;

                string leftStr = (left.TotalDays > 0) ?
                    string.Format("{0:#} {1}:{2:00}:{3:00}", 
                    (int)left.TotalDays, left.Hours, left.Minutes, left.Seconds) : 
                    "past";
                lvItem.SubItems[colLeft.Index].Text = leftStr;
            }
        }

        private void SetSchedule(int idx)
        {
            ListViewItem lvItem = scheduleView.Items[idx];
            while (lvItem.SubItems.Count < scheduleView.Columns.Count)
                lvItem.SubItems.Add("");

            DateTime now = DateTime.Now;
            if (lvItem.Tag == null)
                lvItem.Tag = now;

            TimeSpan elapsed = now - m_startAt;

            DateTime alarmAt = (DateTime)lvItem.Tag;
            TimeSpan left = alarmAt - now;

            string leftStr = (left.TotalDays > 0) ?
                string.Format("{0:#} {1}:{2:00}:{3:00}",
                (int)left.TotalDays, left.Hours, left.Minutes, left.Seconds) : 
                "past";
            string elapsedStr = string.Format("{0:#} {1}:{2:00}:{3:00}",
                elapsed.TotalDays, elapsed.Hours, elapsed.Minutes, elapsed.Seconds);
            string alarmStr = this.m_alarmAt.ToString("dd-MMM-yy HH:mm:ss tt");

            lvItem.SubItems[colTime.Index].Text = alarmStr;
            lvItem.SubItems[colLeft.Index].Text = leftStr;
            lvItem.SubItems[colType.Index].Text = durationBtn.Checked ? s_dur : s_at;
            lvItem.SubItems[colRepeat.Index].Text = repeatCk.Checked ? s_yes : s_no;
            lvItem.SubItems[colSnd.Index].Text = GetFilename(soundTbox.Text);
            lvItem.SubItems[colImage.Index].Text = GetFilename(imageTbox.Text);
            lvItem.SubItems[colText.Index].Text = alarmTbox.Text;
            lvItem.SubItems[colExe.Index].Text = GetFilename(executeTbox.Text);
            lvItem.SubItems[colDur.Index].Text = durTrackBar.Value.ToString();
            lvItem.SubItems[colDurUnit.Index].Text = durUnits.SelectedIndex.ToString();
            m_soundPlayer.Stream = Properties.Resources.chord;
            m_soundPlayer.Play();
        }

        private void AddSchdItem_Click(object sender, EventArgs e)
        {
            scheduleView.SelectedIndices.Clear();
            ListViewItem lvItem = new ListViewItem("?");
            while (lvItem.SubItems.Count < s_subFieldCnt)
                lvItem.SubItems.Add("");
            lvItem.Tag = DateTime.Now + TimeSpan.FromMinutes(5);
            scheduleView.Items.Add(lvItem);
            lvItem.Selected = true;
            Parse();
        }

        private void DelSchdItem_Click(object sender, EventArgs e)
        {
            if (scheduleView.SelectedIndices.Count == 1)
            {
                scheduleView.Items.RemoveAt(scheduleView.SelectedIndices[0]);
                if (scheduleView.Items.Count != 0)
                    scheduleView.Items[scheduleView.Items.Count - 1].Selected = true;
            }
        }

        private void TriggerStateChange()
        {
            if (AlarmStateChanged != null)
                AlarmStateChanged(this, EventArgs.Empty);
        }

        private bool LoadSchedule()
        {
            bool okay = true;
            string[] schdKeys = m_regSchdMgr.GetKeys();
            foreach (string schdKey in schdKeys)
            {
                try
                {
                    DateTime dt = DateTime.Parse(schdKey);
                    string[] fields = m_regSchdMgr.GetString(schdKey).Split(';');
                    ListViewItem lvItem = new ListViewItem();
                    lvItem.Tag = dt;
                    for (int idx = 0; idx != fields.Length; idx++)
                    {
                        if (lvItem.SubItems.Count > idx)
                            lvItem.SubItems[idx].Text = fields[idx];
                        else
                            lvItem.SubItems.Add(fields[idx]);
                    }
                    this.scheduleView.Items.Add(lvItem);
                }
                catch 
                {
                    okay = false;
                }
            }

            TriggerStateChange();
            return okay;
        }

        private void SaveSchedule()
        {
            m_regSchdMgr.DeleteKeys(m_regSchdMgr.GetKeys());

            foreach (ListViewItem lvItem in scheduleView.Items)
            {
                try
                {
                    string fields = string.Empty;
                    foreach (ListViewItem.ListViewSubItem subItem in lvItem.SubItems)
                    {
                        if (fields.Length != 0)
                            fields += ';';
                        fields += subItem.Text;
                    }

                    DateTime dt = (DateTime)lvItem.Tag;
                    m_regSchdMgr.SetString(dt.ToString(), fields);
                }
                catch { }
            }

            m_regSchdMgr.Apply();
        }

        private bool LoadOptions()
        {
            string fontStr = m_regMgr.GetString(s_font);
            if (fontStr != null)
            {
                string[] fontParts = fontStr.Split(';');
                FontStyle fontStyle = (FontStyle)Enum.Parse(typeof(FontStyle), fontParts[2]);
                Font font = new Font(fontParts[0], float.Parse(fontParts[1]), fontStyle);
                fontDialog.Font = font;
            }
            return false;
        }

        private void SaveOptions()
        {
            string fontStr = this.fontDialog.Font.FontFamily.Name + ";";
            fontStr += this.fontDialog.Font.SizeInPoints.ToString() + ";";
            fontStr += this.fontDialog.Font.Style.ToString();
            m_regMgr.SetString(s_font, fontStr);
        }

        class ListViewItemComparer : IComparer
        {
            private int m_column;
            public ListViewItemComparer() { m_column = 0; }
            public ListViewItemComparer(int column) { m_column = column; }
            public int Compare(object x, object y)
            {
                ListViewItem xLv = (ListViewItem)x;
                ListViewItem yLv = (ListViewItem)y;
                DateTime xDt = (DateTime)xLv.Tag;
                DateTime yDt = (DateTime)yLv.Tag;
                return xDt.CompareTo(yDt);
            }
        }

        private void LoadFromSchedule(int idx)
        {
            if (idx < scheduleView.Items.Count)
            {
                ListViewItem lvItem = scheduleView.Items[idx];
                m_alarmAt = (DateTime)lvItem.Tag;
                if (lvItem.SubItems.Count >= 8)
                {
                    m_skipParse = true;
                    durationBtn.Checked = (lvItem.SubItems[colType.Index].Text == s_dur);
                    atTimeBtn.Checked = !durationBtn.Checked;
                    repeatCk.Checked = (lvItem.SubItems[colRepeat.Index].Text == s_yes);
                    soundTbox.Text = lvItem.SubItems[colSnd.Index].Text;
                    imageTbox.Text = lvItem.SubItems[colImage.Index].Text;
                    alarmTbox.Text = lvItem.SubItems[colText.Index].Text;
                    executeTbox.Text = lvItem.SubItems[colExe.Index].Text;
                    if (durationBtn.Checked && lvItem.SubItems.Count >= 10)
                    {
                        int durTrack = int.Parse(lvItem.SubItems[colDur.Index].Text);
                        int durUnit = int.Parse(lvItem.SubItems[colDurUnit.Index].Text);
                        durUnits.SelectedIndex = durUnit;
                        durTrackBar.Value = durTrack;
                    }
                    m_skipParse = false;
                }
            }
        }

        private void scheduleView_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (scheduleView.SelectedIndices.Count != 0)
                LoadFromSchedule(scheduleView.SelectedIndices[0]);
        }

        /// <summary>
        /// Auto stretch view columns to fit dialog size.
        /// </summary>
        private void scheduleView_Resize(object sender, EventArgs e)
        {
            ListView listView = (ListView)sender;
            int colChg = (this.Width - m_prevWidth) / listView.Columns.Count;
            if (m_prevWidth != 0 && colChg != 0)
            {
                foreach (ColumnHeader header in listView.Columns)
                {
                    header.Width += colChg;
                }

                m_prevWidth += colChg * listView.Columns.Count;
            }
            else if (m_prevWidth == 0)
            {
                m_prevWidth = this.Width;
            }
        }
        int m_prevWidth = 0;

        private void splitter1_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button == System.Windows.Forms.MouseButtons.Left)
            {
                int h = scheduleView.PointToClient(System.Windows.Forms.Cursor.Position).Y;
                // scheduleView.Height = h - splitter1.Height;
                panelSchd.Height = (h + panelSchd.Height + 38) / 2;
            }
        }

        private void optionsBtn_Click(object sender, EventArgs e)
        {
            AlarmOptionsDlg optionDlg = new AlarmOptionsDlg(m_optionInfo);
            if (optionDlg.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                m_optionInfo = optionDlg.OptionInfo;
            }
        }

        private void timeLbl_Click(object sender, EventArgs e)
        {
            m_timeFmt = (m_timeFmt + 1) % timeFmts.Length;
        }

        private void alarmAtLbl_Click(object sender, EventArgs e)
        {
            m_alarmFmt = (m_alarmFmt + 1) % timeFmts.Length;
            alarmAtLbl.Text = this.m_alarmAt.ToString(timeFmts[m_alarmFmt]);
        }

        private void animateTimer_Tick(object sender, EventArgs e)
        {
            AnimateNextFrame();
            if ((DateTime.Now - m_alertImageTm).TotalSeconds > 60)
            {
                // Auto close after 60 seconds.
                Alert_Click(sender, e);
            }
        }

        private void alarmTbox_TextChanged(object sender, EventArgs e)
        {
            if (scheduleView.SelectedItems.Count != 0)
            {
                ListViewItem lvItem = scheduleView.SelectedItems[0];
                lvItem.SubItems[colText.Index].Text = alarmTbox.Text;
            }
        }

        private void repeatCk_CheckedChanged(object sender, EventArgs e)
        {
            if (scheduleView.SelectedItems.Count != 0)
            {
                ListViewItem lvItem = scheduleView.SelectedItems[0];
                lvItem.SubItems[colRepeat.Index].Text = repeatCk.Checked ? s_yes : s_no;
            }   
        }

    }
}
