﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace LockDownSvc
{
    /// <summary>
    /// User Interface to edit user configuration values.
    /// 
    /// Author: Dennis Lang
    /// LockDown service - Dec 2010
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public partial class ConfigDlg : Form
    {
        DataMgr m_dataMgr;
        const int userCol = 1;
        const int dateCol = 2;
        const int spanCol = 3;
        const int leftCol = 4;
        const string passwordKeyStr = "string";

        #region ==== Main public interface
        public ConfigDlg(DataMgr dataMgr)
        {
            InitializeComponent();

            configView.ListViewItemSorter = new ListViewColumnSorter(ListViewColumnSorter.SortDataType.eAuto);

            this.Text = Application.ProductName + " v" + Application.ProductVersion;
            m_dataMgr = dataMgr;
            leftError.Text = spanError.Text = string.Empty;

            loadInBackground.RunWorkerAsync();

            string password = Config.InternalDataMgr().GetString(passwordKeyStr);
            if (password != null && password.Length != 0)
            {
                PasswordDlg passwordDlg = new PasswordDlg();
                passwordDlg.Password = password;
                if (passwordDlg.ShowDialog() == System.Windows.Forms.DialogResult.Cancel)
                {
                    this.DialogResult = System.Windows.Forms.DialogResult.Cancel;
                    this.Close();
                    return;
                }
            }

            AddUsers();
        }

        /// <summary>
        /// Fire event if configuration changes for current user.
        /// </summary>
        public EventHandler ConfigurationChanged;
        public class ConfigEvent : EventArgs
        {
            public Config m_cfg;
        }
        #endregion

        /// <summary>
        /// Add all users to UI list.
        /// </summary>
        private void AddUsers()
        {
            string[] users = WinFunc.GetUsers();
            configView.BeginUpdate();
            configView.Items.Clear();
            configView.ListViewItemSorter = null;
            foreach (string user in users)
            {
                ListViewItem lvItem = configView.Items.Add(string.Empty);
                Config config = new Config();
                config.LoadFor(user, m_dataMgr);
                lvItem.Checked = Config.IsAdmin(user);

                lvItem.SubItems.Add(user);
                lvItem.SubItems.Add(config.Date.ToString("yyyy-MMM-dd hh:mm:ss t"));
                lvItem.SubItems.Add(config.SpanString());
                lvItem.SubItems.Add(config.LeftString());
            }
            configView.EndUpdate();
            configView.ListViewItemSorter = new ListViewColumnSorter(ListViewColumnSorter.SortDataType.eAuto);

            applyBtn.Enabled = false;
        }

        /// <summary>
        /// Close UI 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void closeBtn_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        /// <summary>
        /// Return Configuration of selected user.
        /// </summary>
        /// <returns></returns>
        private Config SelectedConfig()
        {
            Config cfg = null;
            if (configView.SelectedItems.Count != 0)
            {
                ListViewItem lvItem = configView.SelectedItems[0];
                cfg = new Config();
                cfg.User = lvItem.SubItems[userCol].Text;
                if (!cfg.ParseSpan(lvItem.SubItems[spanCol].Text) ||
                    !cfg.ParseLeft(lvItem.SubItems[leftCol].Text))
                {
                    return null;
                }
            }

            return cfg;
        }

        /// <summary>
        /// Pulldown Context menu, quick changes to user Span or Left.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void pullDownMenu_Click(object sender, EventArgs e)
        {
            // Check is cursor over time "left" column.
            bool? isLeftCol = null;
            if (configView.Items.Count > 0)
            {
                Rectangle rect4 = configView.Items[0].SubItems[4].Bounds;
                Point pt = this.PointToClient(pullDownMenu.Bounds.Location);
                isLeftCol = rect4.Contains(pt.X, rect4.Y);
            }

            ToolStripDropDownItem dropItem = sender as ToolStripDropDownItem;
            Config cfg = SelectedConfig();
            if (cfg != null)
            {
                ListViewItem lvItem = configView.SelectedItems[0];

                int tagValue = (dropItem != null && dropItem.Tag != null) ? (int)dropItem.Tag : -2;
                switch (tagValue)
                {
                    case -2:    // error - should never get here
                        return;
                    case -1:    // reset
                        cfg.Left = cfg.Span;
                        break;
                    case 0:     // clear
                        cfg.Span = cfg.Left = new TimeSpan(0, 0, 0);
                        break;
                    default:
                        if (isLeftCol != null)
                        {
                            if (isLeftCol == true)
                                cfg.Left = new TimeSpan(0, tagValue, 0);
                            else
                                cfg.Span = new TimeSpan(0, tagValue, 0);
                        }
                        else
                        {
                            cfg.Left = new TimeSpan(0, tagValue, 0);
                            if (cfg.Span < cfg.Left)
                                cfg.Span = cfg.Left;
                        }
                        break;
                    case 999:   // unlimited
                        cfg.SetToUnlimited();
                        break;
                }

                spanBox.Text = lvItem.SubItems[spanCol].Text = cfg.SpanString();
                leftBox.Text = lvItem.SubItems[leftCol].Text = cfg.LeftString();
                ApplyEnable();
            }
        }

        /// <summary>
        /// User selected, populate edit panel and enable Apply button.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void configView_SelectedIndexChanged(object sender, EventArgs e)
        {
            leftError.Text = spanError.Text = string.Empty;

            if (configView.SelectedItems.Count != 0)
            {
                ListViewItem lvItem = configView.SelectedItems[0];

                spanBox.Text = lvItem.SubItems[spanCol].Text;
                leftBox.Text = lvItem.SubItems[leftCol].Text;
                applyBtn.Enabled = true;
            }
            else
            {
                applyBtn.Enabled = false;
            }
        }

        /// <summary>
        /// Column header clicked - fire sort.
        /// </summary>
        private void SortColumn_Click(object sender, ColumnClickEventArgs e)
        {
            ListViewColumnSorter.SortColumn_Click(sender, e);
        }

        /// <summary>
        /// Save changes permentantly.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void applyBtn_Click(object sender, EventArgs e)
        {
            applyBtn.Enabled = false;
            leftError.Text = spanError.Text = string.Empty;

            // First update the administrator list.
            string adminUsers = string.Empty;
            foreach (ListViewItem ckItem in configView.CheckedItems)
            {
                adminUsers += ckItem.SubItems[1].Text + ";";
            }
            Config.AdminUsers = adminUsers;

            // Second, deal with any user configuration changes.
            if (configView.SelectedItems.Count != 0)
            {
                ListViewItem lvItem = configView.SelectedItems[0];
                string user = lvItem.SubItems[userCol].Text;
                string spanText = spanBox.Text;
                string leftText = leftBox.Text;

                Config cfg = new Config();
                bool error = false;
                if (!cfg.ParseSpan(spanText))
                {
                    spanError.Text = "Invalid";
                    error = true;
                }
                if (!cfg.ParseLeft(leftText))
                {
                    leftError.Text = "Invalid";
                    error = true;
                }

                if (!error)
                {
                    cfg.User = user;
                    cfg.SaveFor(user, m_dataMgr);
                    AddUsers();

                    if (ConfigurationChanged != null && user == WinFunc.ForegroundUser)
                        ConfigurationChanged(this, new ConfigEvent() { m_cfg = cfg });
                }
            }
        }

        /// <summary>
        /// Enable 'apply' button, flash button to make it obvious.
        /// </summary>
        private void ApplyEnable()
        {
            if (applyBtn.Enabled == false)
            {
                applyBtn.Enabled = true;

                // Flash apply button.
                Color bgColor = applyBtn.BackColor;
                for (int i = 0; i < 5; i++)
                {
                    applyBtn.BackColor = Color.Red;
                    Application.DoEvents();
                    System.Threading.Thread.Sleep(100);
                    applyBtn.BackColor = bgColor;
                    Application.DoEvents();
                    System.Threading.Thread.Sleep(100);
                }
            }
        }

        /// <summary>
        /// Validate span time.
        /// </summary>
        private void spanBox_KeyUp(object sender, KeyEventArgs e)
        {
            spanError.Text = string.Empty;
            if (e.KeyCode == Keys.Enter)
            {
                Config tstCfg = new Config();
                if (tstCfg.ParseSpan(spanBox.Text) == false)
                    spanError.Text = "Invalid";
            }
        }

        /// <summary>
        /// Validate left time.
        /// </summary>
        private void leftBox_KeyUp(object sender, KeyEventArgs e)
        {
            leftError.Text = string.Empty;
            if (e.KeyCode == Keys.Enter)
            {
                Config tstCfg = new Config();
                if (tstCfg.ParseSpan(leftBox.Text) == false)
                    leftError.Text = "Invalid";
            }
        }

        /// <summary>
        /// Auto stretch view columns to fit dialog size.
        /// </summary>
        private void configView_Resize(object sender, EventArgs e)
        {
            int colChg = (this.Width - m_prevWidth) / configView.Columns.Count;
            if (m_prevWidth != 0 && colChg != 0)
            {
                foreach (ColumnHeader header in configView.Columns)
                {
                    header.Width += colChg;
                }

                m_prevWidth += colChg * configView.Columns.Count;
            }
            else if (m_prevWidth == 0)
            {
                m_prevWidth = this.Width;
            }
        }
        int m_prevWidth = 0;

        private void loadInBackground_DoWork(object sender, DoWorkEventArgs e)
        {
            // This can take a long time, the first time it is called.
            string[] users = WinFunc.GetUsers();
        }

        private void loadInBackground_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            busyLabel.Visible = false;
            WinFunc.GetUsers(); // first time callled this can be slow. Results are cached.
        }

        #region ==== Password
        private void passwordBtn_Click(object sender, EventArgs e)
        {
            PasswordDlg passwordDlg = new PasswordDlg();
            passwordDlg.Password = Config.InternalDataMgr().GetString(passwordKeyStr);
            passwordDlg.Mode = PasswordDlg.PasswordMode.eAccept;
            if (passwordDlg.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                DataMgr internalDataMgr = Config.InternalDataMgr();
                internalDataMgr.SetString(passwordKeyStr,passwordDlg.Password);
                internalDataMgr.Apply();
            }
        }

   
        #endregion

    }
}
