﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Media;

namespace LockDownSvc 
{
    /// <summary>
    /// Display Detail User Time History
    /// 
    /// Author: Dennis Lang
    /// LockDown service - Dec 2010
    /// http://home.comcast.net/~lang.dennis/
    /// </summary>
    public partial class HistoryDlg : Form
    {
        private LockDownFg m_lockDown;

        public HistoryDlg(LockDownFg lockDownFg)
        {
            InitializeComponent();

            historyView.ListViewItemSorter = new ListViewColumnSorter(ListViewColumnSorter.SortDataType.eAuto);
            exportFileDialog.InitialDirectory = System.Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            m_lockDown = lockDownFg;
            LoadHistory();
        }

        private void closeBtn_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        /// <summary>
        /// Historical time categories.
        /// </summary>
        class UserHist
        {
            public TimeSpan today = TimeSpan.Zero;
            public TimeSpan yesterday = TimeSpan.Zero;
            public TimeSpan n7Days = TimeSpan.Zero;
            public TimeSpan total = TimeSpan.Zero;
            public int dayCnt = 0;
            public int day = -1;
        }

        /// <summary>
        /// Load history detail from log file and pouplate history view.
        /// </summary>
        public void LoadHistory()
        {
            historyView.Items.Clear();
            status.Text = string.Empty;

            try
            {
                // Store details by user.
                Dictionary<string, UserHist> histDetail = new Dictionary<string, UserHist>();

                // Time category thresholds.
                DateTime now = DateTime.Now;
                DateTime beginOfToday = new DateTime(now.Year, now.Month, now.Day, 0, 0, 0);
                DateTime beginOfYesterday = beginOfToday - TimeSpan.FromDays(1);
                DateTime beginOf7Day = beginOfToday - TimeSpan.FromDays(6);

                // Parse log file.
                StreamReader logReader = File.OpenText(Log.LogFile);
                string line;

                while ((line = logReader.ReadLine()) != null)
                {
                    // Sample log file:
                    // Syntax:
                    //     ShortDate HH:MM:SS, [user], action, msg
                    // Where msg is:
                    //     msg = "User=<user>, Date=<date>, Span=<span>, Left=<left>, Elapsed=<elapsed>
                    // Example:
                    // 12/28/2010 19:12,[dennis],LogOff,User=none Date=12/28/2010 6:13:56 PM Span=2:00:00 Left=0:0:0 Elapsed=0:00:05

                    try
                    {
                        string[] fields = line.Split(',');
                        if (fields.Length >= 8 && fields[2].ToLower().Contains("logoff"))
                        {
                            string user = fields[1].Trim('[').Trim(']');

                            // Locate lines with elapsed time.
                            int idx = fields[7].ToLower().IndexOf("elapsed=");
                            if (idx != -1)
                            {
                                string elapsedStr = fields[7].Substring(idx + 8);
                                TimeSpan elapsed = TimeSpan.Parse(elapsedStr);
                                DateTime date = DateTime.Parse(fields[0]);

                                if (!histDetail.ContainsKey(user))
                                    histDetail.Add(user, new UserHist());

                                UserHist userHist = histDetail[user];

                                if (date.Day != userHist.day)
                                {
                                    userHist.day = date.Day;
                                    userHist.dayCnt++;
                                }
                                userHist.total += elapsed;

                                if (date >= beginOfToday)
                                    userHist.today += elapsed;
                                else if (date >= beginOfYesterday)
                                    userHist.yesterday += elapsed;

                                if (date >= beginOf7Day)
                                    userHist.n7Days += elapsed;
                            }
                        }
                    }
                    catch { }   // parsing error.
                }
                logReader.Close();

                // Add in current time.
                if (!histDetail.ContainsKey(m_lockDown.Configuration.User))
                    histDetail.Add(m_lockDown.Configuration.User, new UserHist());
                histDetail[m_lockDown.Configuration.User].today += m_lockDown.Configuration.Elapsed;
                histDetail[m_lockDown.Configuration.User].n7Days += m_lockDown.Configuration.Elapsed; 

                // Populate history view.
                historyView.BeginUpdate();
                foreach (string user in histDetail.Keys)
                {
                    UserHist userHist = histDetail[user];

                    ListViewItem lvItem = this.historyView.Items.Add(user);
                    lvItem.SubItems.Add(FormatTs(userHist.today));
                    lvItem.SubItems.Add(FormatTs(userHist.yesterday));
                    lvItem.SubItems.Add(FormatTs(userHist.n7Days));
                    lvItem.SubItems.Add(FormatTs(userHist.total));

                    lvItem.SubItems.Add(userHist.dayCnt.ToString("N0"));
                    double dailyAvg = (userHist.dayCnt > 0) ? userHist.total.TotalHours / userHist.dayCnt : 0;
                    lvItem.SubItems.Add(dailyAvg.ToString("N2"));
                }
                historyView.EndUpdate();
            }
            catch (Exception ex)
            {
                Log.LogEvent("History load failed " + ex.Message);
            }
        }

        private void exportBtn_Click(object sender, EventArgs e)
        {
            if (exportFileDialog.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                try
                {
                    StreamWriter writer = File.CreateText(exportFileDialog.FileName);

                    string line = string.Empty;
                    foreach (ColumnHeader hdr in historyView.Columns)
                    {
                        if (line.Length != 0)
                            line += ",";
                        line += hdr.Text;
                    }
                    writer.WriteLine(line);

                    foreach (ListViewItem lvItem in historyView.Items)
                    {
                        line = string.Empty;
                        foreach (ListViewItem.ListViewSubItem subItem in lvItem.SubItems)
                        {
                            if (line.Length != 0)
                                line += ",";
                            line += subItem.Text;
                        }
                        writer.WriteLine(line);
                    }
                    writer.Close();
                    status.Text = "Exported completed";
                }
                catch (Exception ex)
                {
                    status.Text = "Failed to export - " + ex.Message;
                    SoundPlayer player = new SoundPlayer();
                    player.Stream = Properties.Resources.chord;
                    player.Play();
                }
            }
        }

        /// <summary>
        /// Column header clicked - fire sort.
        /// </summary>
        private void SortColumn_Click(object sender, ColumnClickEventArgs e)
        {
            ListViewColumnSorter.SortColumn_Click(sender, e);
        }

        bool m_formatHHMM = false;
        private void formatBtn_Click(object sender, EventArgs e)
        {
            m_formatHHMM = !m_formatHHMM;
            formatBtn.Text = m_formatHHMM ? "HH.fff" : "HH:MM";
            LoadHistory();
        }

        private string FormatTs(TimeSpan ts)
        {
            if (m_formatHHMM)
                return string.Format("{0:#} {1:00}:{2:00}", (int)ts.TotalDays, ts.Hours, ts.Minutes);
            else
                return string.Format("{0:N2}", ts.TotalHours);
        }

        /// <summary>
        /// Auto stretch view columns to fit dialog size.
        /// </summary>
        private void historyView_Resize(object sender, EventArgs e)
        {
            ListView listView = (ListView)sender;
            int colChg = (this.Width - m_prevWidth) / listView.Columns.Count;
            if (m_prevWidth != 0 && colChg != 0)
            {
                foreach (ColumnHeader header in listView.Columns)
                {
                    header.Width += colChg;
                }

                m_prevWidth += colChg * listView.Columns.Count;
            }
            else if (m_prevWidth == 0)
            {
                m_prevWidth = this.Width;
            }
        }
        int m_prevWidth = 0;
    }
}
